# solver.py
# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

"""
This module is the central module of the analysis phase.
Based on the given entries in settings and static_settings, the functions solver and adaptive_load_stepping can be used to find the roots of the global residual vector. 
Depending on the settings, linear equation solvers, the Newton-Raphson method, or nonlinear minimizers are utilized. 
The residual vectors and (in the case of external solvers) the tangent matrix are automatically assembled according to the chosen settings. 
The solver module uses the assembler, which in turn calls the variational_scheme, the solution_structure, and the spaces modules. 
Additionally, automatic implicit differentiation in forward or reverse mode via the implicit_diff module is provided for the adaptive_load_stepping function.
For solving the linear equation systems, wrappers for different backends on CPU and GPU are available, including Pardiso and PETSc.
"""

import time
import sys

import jax
import jaxopt
import jax.numpy as jnp
from jax.experimental import sparse
from jax import lax
import numpy as np
import scipy as scp
from flax.core import FrozenDict

from autopdex import assembler, implicit_diff, utility



### Solvers as specified by the static_settings and settings
@utility.jit_with_docstring(static_argnames=["static_settings", "**kwargs"])
def solver(dofs, settings, static_settings, **kwargs):
    """
    General solver function to solve a given problem based on provided settings.

    This function chooses and runs the appropriate solver type (e.g., minimization, linear, Newton) based on the provided
    `static_settings` and returns the solution and any additional information.

    Args:
        dofs (jnp.ndarray or dict): Degrees of freedom or initial guess for the solution.
        settings (dict): Dictionary containing various settings and parameters required for assembling the problem.
        static_settings (dict): Dictionary containing static settings such as solver type, verbose level, and variational schemes.
        **kwargs (dict): Additional keyword arguments passed to the specific solver functions.

    Returns:
        jnp.ndarray: The solution obtained from the selected solver.
        Any: Additional information from the solver, such as number of iterations or convergence status.

    Solver Types:
      - 'minimize' : Uses nonlinear minimization solvers (e.g., LBFGS, BFGS, etc.).
      - 'linear' : Solves linear systems using specified backend (e.g., JAX, PETSc, PARDISO, PyAMG, Scipy).
      - 'diagonal linear' : Solves linear systems assuming a diagonal tangent matrix.
      - 'newton' : Uses the Newton method for solving nonlinear systems.
      - 'damped newton' : Uses a damped Newton method for solving nonlinear systems.

    For the different solvers, the function conducts the following functions in which more documentation is provided:
      - 'minimize' : solver.solve_nonlinear_minimization
      - 'linear' : solver.solve_linear
      - 'diagonal linear' : solver.solve_diagonal_linear
      - 'newton' : solver.solve_newton
      - 'damped newton' : solver.solve_damped_newton

    Notes:
      - If all domains are using 'least square pde loss' variational scheme and verbosity level is >=1, it prints the L2 error before and after optimization.
    """

    # Give global error estimator with unfitted dofs if all domains are 'least square pde loss'
    verbose = static_settings["verbose"]
    try:
        if (
            all(
                [
                    scheme == "least square pde loss"
                    for scheme in static_settings["variational scheme"]
                ]
            )
            and verbose >= 1
        ):
            jax.debug.print(
                "L2 error unoptimized: {x}",
                x=assembler.integrate_functional(dofs, settings, static_settings),
            )
    except KeyError:
        pass

    # Choose type of solver and call it
    solver_type = static_settings["solver type"]
    match solver_type:
        case "minimize":
            sol = solve_nonlinear_minimization(
                dofs, settings, static_settings, **kwargs
            )
            infos = None
        case "linear":
            sol = solve_linear(dofs, settings, static_settings, **kwargs)
            infos = None
        case "diagonal linear":
            sol = solve_diagonal_linear(dofs, settings, static_settings, **kwargs)
            infos = None
        # case 'diagonal newton':
        #   # ToDo
        #   sol, infos = solve_diagonal_newton(dofs, settings, static_settings, **kwargs)
        case "newton":
            sol, infos = solve_newton(dofs, settings, static_settings, **kwargs)
        case "damped newton":
            sol, infos = solve_damped_newton(dofs, settings, static_settings, **kwargs)
        case _:
            assert False, "Solver type not implemented!"

    # Give global error estimator with fitted dofs if all domains are 'least square pde loss'
    try:
        if (
            all(
                [
                    scheme == "least square pde loss"
                    for scheme in static_settings["variational scheme"]
                ]
            )
            and verbose >= 1
        ):
            jax.debug.print(
                "L2 error optimized: {x}",
                x=assembler.integrate_functional(sol, settings, static_settings),
            )
    except KeyError:
        pass
    return sol, infos

@utility.jit_with_docstring(
    static_argnames=[
        "static_settings",
        "multiplier_settings",
        "path_dependent",
        "implicit_diff_mode",
        "max_load_steps",
        "max_multiplier",
        "min_increment",
        "max_increment",
        "init_increment",
        "target_num_newton_iter",
        "newton_tol",
        "**kwargs",
    ]
    )
def adaptive_load_stepping(
    dofs,
    settings,
    static_settings,
    multiplier_settings=lambda settings, multiplier: (
        settings.update({"load multiplier": multiplier}),
        settings,
    )[1],
    path_dependent=True,
    implicit_diff_mode=None,
    max_multiplier=1.0,
    min_increment=0.01,
    max_increment=1.0,
    init_increment=0.2,
    max_load_steps=1000,
    target_num_newton_iter=7,
    newton_tol=1e-10,
    **kwargs,
    ):
    """
    Performs adaptive load stepping to solve a nonlinear system of equations.

    This function iteratively adjusts the load increment to ensure convergence
    using a Newton-Raphson solver. The increment size is adaptively controlled
    based on the convergence behavior of the solver. Works currently only with
    solver types 'newton' and 'damped newton'.

    Args:
      dofs (jnp.ndarray or dict): Initial degrees of freedom.
      settings (dict): Dictionary of problem settings.
      static_settings (dict): Dictionary of static settings that do not change during load steps.
      multiplier_settings (callable): Function to update settings based on the current load multiplier.
      path_dependent (bool): Specifies wether problem is path-dependent (experimental) or not (has an influence on the implicit differentiation).
      implicit_diff_mode (string): Can be either \'reverse\', \'forward\' or None. In case of \'reverse\', only reverse mode differentiation is supported (jacrev), in case of \'forward\', only forward mode differentiation is supported (jacfwd).
      max_multiplier (float): Maximum value for the load multiplier.
      min_increment (float): Minimum allowable increment size.
      max_increment (float): Maximum allowable increment size.
      init_increment (float): Initial increment size.
      max_load_steps (int): Maximal number of load steps. Only used in case implicit_diff_mod is not None
      target_num_newton_iter (int): Target number of Newton iterations for each load step.
      newton_tol (float, optional): Tolerance for Newton solver convergence. Default is 1e-10.
      **kwargs: Additional keyword arguments for the solver.

    Returns:
        - jnp.ndarray: Solution degrees of freedom after load stepping.
    """
    verbose = static_settings["verbose"]

    if implicit_diff_mode is not None:
        # Set-up the decorator for implicit differentiation
        residual_fun = lambda dofs, settings: assembler.assemble_residual(dofs, settings, static_settings)
        tangent_fun = lambda dofs, settings: assembler.assemble_tangent(dofs, settings, static_settings)

        try:
            dirichlet_dofs = settings["dirichlet dofs"]
        except KeyError:
            # Warning if it was defined in static_settings
            assert "dirichlet dofs" not in static_settings, \
                "'dirichlet dofs' has been moved to 'settings' in order to reduce compile time. \
                Further, you should not transform it to a tuple of tuples anymore."
            pass
        free_dofs = None
        free_dofs_flat = None
        if dirichlet_dofs is not None:
            free_dofs = utility.mask_op(dirichlet_dofs, utility.dict_ones_like(dirichlet_dofs), mode="apply", ufunc=lambda x: ~x)
            free_dofs_flat = utility.dict_flatten(free_dofs)
        is_constrained = True if free_dofs is not None else False

        solver_backend = static_settings["solver backend"]
        solver_subtype = static_settings["solver"]
        try:
            sensitivity_solver_backend = static_settings["sensitivity solver backend"]
            sensitivity_solver_subtype = static_settings["sensitivity solver"]
        except KeyError:
            sensitivity_solver_backend = solver_backend
            sensitivity_solver_subtype = solver_subtype

        match sensitivity_solver_backend:
            case "petsc":
                n_fields = static_settings["number of fields"]
                pc_type = static_settings["type of preconditioner"]
                lin_solve_fun = lambda mat, rhs, free_dofs_flat: linear_solve_petsc(
                    mat,
                    rhs,
                    n_fields,
                    sensitivity_solver_subtype,
                    pc_type,
                    verbose,
                    free_dofs_flat,
                    **kwargs,
                )
            case "pardiso":
                lin_solve_fun = lambda mat, rhs, free_dofs_flat: linear_solve_pardiso(
                    mat,
                    rhs,
                    sensitivity_solver_subtype,
                    verbose,
                    free_dofs_flat,
                    **kwargs,
                )
            case "pyamg":
                pc_type = static_settings["type of preconditioner"]
                lin_solve_fun = lambda mat, rhs, free_dofs_flat: linear_solve_pyamg(
                    mat,
                    rhs,
                    sensitivity_solver_subtype,
                    pc_type,
                    verbose,
                    free_dofs_flat,
                    **kwargs,
                )
            case "scipy":
                lin_solve_fun = lambda mat, rhs, free_dofs_flat: linear_solve_scipy(
                    mat,
                    rhs,
                    sensitivity_solver_subtype,
                    verbose,
                    free_dofs_flat,
                    **kwargs,
                )
            case _:
                raise ValueError(
                    "Specified sensitivity solver backend not available. Choose 'pardiso', 'petsc', 'pyamg' or 'scipy'."
                )

        def lin_solve_callback_fun(mat, rhs, free_dofs_flat):
            rhs_flat = utility.dict_flatten(rhs)
            sol = jax.pure_callback(lin_solve_fun, jnp.zeros(rhs_flat.shape, rhs_flat.dtype), mat, rhs_flat, free_dofs_flat, vmap_method='sequential')
            return utility.reshape_as(sol, rhs)

        # def lin_solve_callback_fun(mat, rhs, free_dofs_flat):
        #     rhs_flat = utility.dict_flatten(rhs)
        #     # linear_solver = lambda mat, rhs: jax.pure_callback(lin_solve_fun, jnp.zeros(rhs_flat.shape, rhs_flat.dtype), mat, rhs_flat, None, vmap_method='sequential')
        #     linear_solver = lambda mat, rhs: jax.pure_callback(lin_solve_fun, jnp.zeros(rhs_flat.shape, rhs_flat.dtype), mat, rhs_flat, free_dofs_flat, vmap_method='sequential')
        #     solve = lambda matvec, x: linear_solver(mat, x)
        #     transpose_solve = lambda vecmat, x: linear_solver(mat.T, x)
        #     matvec = lambda x: utility.reshape_as(mat @ utility.dict_flatten(x), x)
        #     sol = jax.lax.custom_linear_solve(matvec, rhs_flat, solve, transpose_solve)
        #     return utility.reshape_as(sol, rhs)

    # Set up functions for adaptive load stepping loop
    def continue_check(carry):
        _, multiplier, increment, _, _, _ = carry
        _continue_1 = jnp.logical_and(
            multiplier < max_multiplier, increment > min_increment
        )
        jax.lax.cond(
            jnp.logical_and(
                multiplier < max_multiplier - min_increment, increment < min_increment
            ),
            lambda: jax.debug.print(
                "Adaptive load stepping could not converge; increment size below min_increment!"
            ),
            lambda: None,
        )
        return _continue_1

    def step(carry):
        dofs0, multiplier, increment, load_step, settings, _ = carry

        # Update multiplier
        multiplier += increment

        # Update boundary conditions
        if verbose > -1:
            if verbose > 0:
                jax.debug.print("")
            jax.debug.print("Multiplier: {x}", x=multiplier)
        settings = multiplier_settings(settings, multiplier)

        # Call newton solver
        if (
            path_dependent and implicit_diff_mode is not None
        ):  # Add implicit differentiation for each load step

            @implicit_diff.custom_root(
                residual_fun,
                tangent_fun,
                lin_solve_callback_fun,
                is_constrained,
                True,
                implicit_diff_mode,
            )
            def diffable_solve(dofs0, settings):
                dofs, (needed_steps, res_norm_free_dofs, divergence) = solver(
                    dofs0, settings, static_settings, newton_tol=newton_tol, **kwargs
                )
                return dofs, (needed_steps.astype(float), res_norm_free_dofs, divergence.astype(float))

            dofs, infos = diffable_solve(dofs0, settings)
            needed_steps, res_norm_free_dofs, divergence = infos
            divergence = divergence.astype(bool)


        else:  # Add implicit diff wrapper on the adaptive load stepping level
            dofs, infos = solver(
                dofs0, settings, static_settings, newton_tol=newton_tol, **kwargs
            )
            needed_steps, res_norm_free_dofs, divergence = infos

        # Adaptive incrementation
        multiplier = jnp.where(divergence, multiplier - increment, multiplier)
        increment = jnp.where(
            divergence,
            0.5 * increment,
            (1 + 0.5 * (target_num_newton_iter - needed_steps) / target_num_newton_iter)
            * increment,
        )
        increment = jnp.where(increment > max_increment, max_increment, increment)

        if isinstance(dofs, dict):
            dofs = {
                key: jnp.where(divergence, dofs0[key], dofs[key])
                for key in dofs0.keys()
            }
        else:
            dofs = jnp.where(divergence, dofs0, dofs)

        # Limit to max_multiplier
        increment = jnp.where(
            multiplier + increment > max_multiplier,
            max_multiplier - multiplier,
            increment,
        )
        return (dofs, multiplier, increment, 1.0 * load_step, settings, res_norm_free_dofs)

    # Use implicit diff wrappers to make it differentiable
    if implicit_diff_mode is not None:
        if not path_dependent:  # Conservative problem

            # Set Dirichlet conditions for derivative w.r.t. them
            settings = multiplier_settings(settings, max_multiplier)

            @implicit_diff.custom_root(
                residual_fun,
                tangent_fun,
                lin_solve_callback_fun,
                is_constrained,
                True,
                implicit_diff_mode,
            )
            def diffable_adaptive_load_stepping(dofs, settings):
                (dofs, multiplier, increment, load_step, settings, res_norm_free_dofs) = \
                    jax.lax.while_loop(
                        continue_check, step, (dofs, 0.0, init_increment, 0, settings, 0.0)
                        )
                return dofs, (multiplier, increment, load_step, res_norm_free_dofs)

            dofs, (multiplier, increment, load_step, res_norm_free_dofs) = \
                diffable_adaptive_load_stepping(dofs, settings)
            return dofs, (multiplier, increment, load_step, settings, res_norm_free_dofs)

        else:  # Pathdependent problem; uses fori_loop with static limits for supporting reverse mode differentiation

            # Set Dirichlet conditions for derivative w.r.t. them
            settings = multiplier_settings(settings, max_multiplier)

            def body_fn(i, carry):
                def continue_check(carry):
                    _, multiplier, increment, _, _, divergence, stop = carry
                    _continue_1 = jnp.logical_and(
                        multiplier < max_multiplier, increment > min_increment
                    )
                    jax.lax.cond(
                        jnp.logical_and(
                            jnp.logical_and(
                                multiplier < max_multiplier, increment < min_increment
                            ),
                            jnp.logical_not(stop),
                        ),
                        lambda: jax.debug.print(
                            "Adaptive load stepping could not converge; increment size below min_increment!"
                        ),
                        lambda: None,
                    )
                    return _continue_1

                def step_extended(carry):
                    (dofs, multiplier, increment, load_step, settings, res_norm_free_dofs, stop) = carry
                    args = (dofs, multiplier, increment, load_step, settings, res_norm_free_dofs)
                    return step(args) + (False,)

                def finish(carry):
                    (dofs, multiplier, increment, load_step, settings, res_norm_free_dofs, stop) = carry
                    return (dofs, multiplier, increment, load_step, settings, res_norm_free_dofs, True)

                carry = jax.lax.cond(
                    continue_check(carry),
                    lambda x: step_extended(x),
                    lambda x: finish(x),
                    carry,
                )

                # ToDo: Verify accuracy of derivatives with finite differences.
                return carry

            init_state = (dofs, 0.0, init_increment, 0., settings, 0.0, False)
            return jax.lax.fori_loop(0, max_load_steps, body_fn, init_state)

    else:  # No definition of implicit derivatives
        return jax.lax.while_loop(
            continue_check, step, (dofs, 0.0, init_increment, 0., settings, 0.0)
        )

### Minimizers
@utility.jit_with_docstring(static_argnames=["static_settings", "**kwargs"])
def solve_nonlinear_minimization(dofs, settings, static_settings, **kwargs):
    """
    Solves a nonlinear minimization problem using specified optimization methods.

    This function wraps nonlinear minimization solvers provided by `jaxopt` to minimize a functional
    and solve the given problem.

    Args:
      dofs (jnp.ndarray or dict): Degrees of freedom or initial guess for the solution.
      settings (dict): Dictionary containing various settings and parameters required for assembling the problem.
      static_settings (dict): Dictionary containing static settings such as solver type, verbose level, and variational schemes.
      **kwargs (dict): Additional keyword arguments passed to the specific solver functions.

    Returns:
      jnp.ndarray: The optimized solution obtained from the selected solver.

    Solver Types
      - 'gradient descent' : Uses gradient descent for optimization.
      - 'lbfgs' : Uses Limited-memory Broyden-Fletcher-Goldfarb-Shanno (LBFGS) algorithm for optimization.
      - 'bfgs' : Uses Broyden-Fletcher-Goldfarb-Shanno (BFGS) algorithm for optimization.
      - 'nonlinear cg' : Uses nonlinear conjugate gradient method for optimization.
      - 'gauss newton' : Uses Gauss-Newton method for optimization.
      - 'levenberg marquart' : Uses Levenberg-Marquardt algorithm for optimization.
      - Default : If solver name is not set or not available, uses 'lbfgs' as the default solver.

    Notes:
      - This function should just be called, if the variational scheme involves the definition of a functional that is to be minimized,
        e.g. 'least square pde loss'. The modes 'gauss newton' and 'levenberg marquart' are an exeption, since they utilize the residual.
      - The function conducts the assembler.integrate_functional and assembler.assemble_residual functions in order
        to set up suitable optimization functions or residual functions, depending on what the solver needs.
      - The current implementation does not support nodal imposition of DOFs.
    """
    nodal_imposition = "nodal imposition" in static_settings["solution structure"]
    assert (
        not nodal_imposition
    ), "solver type 'minimize' does currently not support nodal imposition of DOFs."
    # ToDo: impose boundary conditions and freeze dirichlet dofs

    def functional(params):
        return assembler.integrate_functional(params, settings, static_settings)

    def residual_function(params):
        return assembler.assemble_residual(params, settings, static_settings)

    # Select minimizer
    solver_name = static_settings["solver"]
    match solver_name:
        case "gradient descent":
            solver = jaxopt.GradientDescent(functional, **kwargs)
            sol = solver.run(dofs).params
        case "lbfgs":
            solver = jaxopt.LBFGS(functional, **kwargs)
            sol = solver.run(dofs).params
        case "bfgs":
            solver = jaxopt.BFGS(functional, **kwargs)
            sol = solver.run(dofs).params
        case "nonlinear cg":
            solver = jaxopt.NonlinearCG(functional, **kwargs)
            sol = solver.run(dofs).params
        case "gauss newton":
            solver = jaxopt.GaussNewton(residual_function, **kwargs)
            sol = solver.run(dofs).params
        case "levenberg marquart":
            assert not isinstance(
                dofs, dict
            ), "solver 'levenberg marquart' does currently not support dicts as dofs"

            def residual_function_flat(params):
                return assembler.assemble_residual(
                    jnp.reshape(params, dofs.shape), settings, static_settings
                ).flatten()

            solver = jaxopt.LevenbergMarquardt(residual_function_flat, **kwargs)
            sol = jnp.reshape(solver.run(dofs.flatten()).params, dofs.shape)
        case _:
            solver = jaxopt.LBFGS(functional, **kwargs)
            sol = solver.run(dofs).params
            print(
                "Solver name not set or not available in combination with this solver type. Using static_settings['solver name'] = 'lbfgs' as default."
            )
    return sol

### Root finders
@utility.jit_with_docstring(static_argnames=["static_settings", "**kwargs"])
def solve_linear(dofs, settings, static_settings, **kwargs):
    """
    Solves a linear system using the specified backend and solver settings.

    This function determines the appropriate linear solver based on the provided settings and forwards the
    call to the selected solver function. It supports both JAX matrix-free solvers and external solvers
    like PETSc, PARDISO, PyAMG, and Scipy using jax.pure_callback.

    Args:
      dofs (jnp.ndarray or dict): Degrees of freedom or initial guess for the solution.
      settings (dict): Dictionary containing various settings and parameters required for assembling the problem.
      static_settings (dict): Dictionary containing static settings such as solver backend, type of solver, and preconditioner.
      **kwargs (dict): Additional keyword arguments passed to the specific solver functions.

    Returns:
      jnp.ndarray: The solution obtained from the selected linear solver.

    Solver Backends:
      - 'jax' : Uses JAX's matrix-free solvers.
      - 'petsc' : Uses PETSc for solving linear systems.
      - 'pardiso' : Uses PARDISO for solving linear systems.
      - 'pyamg' : Uses PyAMG for solving linear systems.
      - 'scipy' : Uses Scipy's sparse solvers for solving linear systems.

    Notes:
      - If `nodal imposition` is detected in the `static_settings`, the function imposes Dirichlet boundary conditions
        and adjusts the degrees of freedom accordingly.
      - The function assembles the residual and tangent matrix before solving the system in case an external solver is used.
      - If the tangent matrix is dense and an external solver is used, it is converted to a sparse format.
      - The function uses JAX's `pure_callback` to call external solvers and handle the solution.
    """
    solver_backend = static_settings["solver backend"]

    ### Jax matrix-free solver
    if solver_backend == "jax":
        return linear_solve_jax(dofs, settings, static_settings, **kwargs)

    ### External linear solver

    nodal_imposition = "nodal imposition" in static_settings["solution structure"]
    # Impose nodal dofs
    if nodal_imposition:
        dirichlet_conditions = settings["dirichlet conditions"]

        if isinstance(settings["dirichlet dofs"], (dict, FrozenDict)):
            dirichlet_dofs_dict_flat = {
                key: jnp.asarray(val).flatten()
                for key, val in settings["dirichlet dofs"].items()
            }
            dirichlet_dofs_flat = jnp.concatenate(
                list(dirichlet_dofs_dict_flat.values())
            )
            dofs = utility.mask_op(dofs, dirichlet_dofs_dict_flat, dirichlet_conditions)
        else:
            dirichlet_dofs_flat = jnp.asarray(
                settings["dirichlet dofs"]
            ).flatten()
            dofs = utility.mask_op(dofs, dirichlet_dofs_flat, dirichlet_conditions)

        free_dofs_flat = jnp.invert(dirichlet_dofs_flat)

    # Assembling
    verbose = static_settings["verbose"]
    solver = static_settings["solver"]

    rhs = assembler.assemble_residual(dofs, settings, static_settings)
    rhs = -utility.dict_flatten(rhs)
    mat = assembler.assemble_tangent(dofs, settings, static_settings)

    # If dense matrix and external solver, convert to sparse
    if (
        solver_backend in ("petsc", "pardiso", "pyamg", "scipy")
        and type(mat) == jnp.ndarray
    ):
        mat = sparse.bcoo_fromdense()

    match solver_backend:
        case "petsc":
            n_fields = static_settings["number of fields"]
            pc_type = static_settings["type of preconditioner"]
            solve_fun = lambda a, b, c: linear_solve_petsc(
                a, b, n_fields, solver, pc_type, verbose, free_dofs=c, **kwargs
            )
        case "pardiso":
            solve_fun = lambda a, b, c: linear_solve_pardiso(
                a, b, solver, verbose, free_dofs=c
            )
        case "pyamg":
            pc_type = static_settings["type of preconditioner"]
            solve_fun = lambda a, b, c: linear_solve_pyamg(
                a, b, solver, pc_type, verbose, free_dofs=c, **kwargs
            )
        case "scipy":
            solve_fun = lambda a, b, c: linear_solve_scipy(
                a, b, solver, verbose, free_dofs=c
            )

    # Prepare callback
    result_shape_dtype = jax.ShapeDtypeStruct(shape=rhs.shape, dtype=rhs.dtype)

    # Compose solution dofs
    if nodal_imposition:
        sol = jax.pure_callback(solve_fun, result_shape_dtype, mat, rhs, free_dofs_flat, vmap_method='sequential')
        if isinstance(settings["dirichlet dofs"], (dict, FrozenDict)):
            free_dofs_dict = {
                key: jnp.invert(jnp.asarray(val))
                for key, val in settings["dirichlet dofs"].items()
            }
            sol = utility.reshape_as(sol, dofs)
            return utility.mask_op(dofs, free_dofs_dict, sol)
        else:
            return utility.mask_op(dofs, free_dofs_flat, sol)
    else:
        sol = jax.pure_callback(solve_fun, result_shape_dtype, mat, rhs, None, vmap_method='sequential')
        return utility.reshape_as(sol, dofs)

@utility.jit_with_docstring(static_argnames=["static_settings", "**kwargs"])
def solve_diagonal_linear(dofs, settings, static_settings, **kwargs):
    """
    Solves a linear system assuming the tangent matrix is diagonal.

    This function solves the linear system by leveraging the assumption that the tangent matrix is diagonal,
    which simplifies the solution process. It supports nodal imposition of Dirichlet boundary conditions and
    handles the assembly of the residual and diagonal tangent matrix.

    If the tangent matrix is not diagonal, it will produce a wrong diagonal of the tangent!

    Args:
      dofs (jnp.ndarray or dict): Degrees of freedom or initial guess for the solution.
      settings (dict): Dictionary containing various settings and parameters required for assembling the problem.
      static_settings (dict): Dictionary containing static settings such as solution structure and solver backend.
      **kwargs (dict): Additional keyword arguments passed to the solver function.

    Returns:
      sol (jnp.ndarray): The solution obtained by solving the linear system assuming a diagonal tangent matrix.

    Notes:
      - If `nodal imposition` is detected in the `static_settings`, the function imposes Dirichlet boundary conditions
        and adjusts the degrees of freedom accordingly.
      - The function assembles the residual and diagonal tangent matrix before solving the system.
      - The solution process involves element-wise division of the residual by the diagonal elements of the tangent matrix.
    """
    nodal_imposition = "nodal imposition" in static_settings["solution structure"]

    # Impose nodal dofs
    if nodal_imposition:
        dirichlet_conditions = utility.dict_flatten(settings["dirichlet conditions"])

        if isinstance(settings["dirichlet dofs"], (dict, FrozenDict)):
            dirichlet_dofs_dict_flat = {
                key: jnp.asarray(val).flatten()
                for key, val in settings["dirichlet dofs"].items()
            }
            dirichlet_dofs_flat = jnp.concatenate(
                list(dirichlet_dofs_dict_flat.vlaues())
            )
            dofs = utility.mask_op(
                dofs, dirichlet_dofs_dict_flat, dirichlet_conditions
            )
        else:
            dirichlet_dofs_flat = jnp.asarray(
                settings["dirichlet dofs"]
            ).flatten()
            dofs = utility.mask_op(dofs, dirichlet_dofs_flat, dirichlet_conditions)

        free_dofs_flat = jnp.invert(dirichlet_dofs_flat)

    # Assembling
    rhs = -assembler.assemble_residual(dofs, settings, static_settings)
    diag_mat = assembler.assemble_tangent(dofs, settings, static_settings).data

    # Delete rows
    rhs = utility.dict_flatten(rhs)
    if nodal_imposition:
        # Delete rows
        rhs = rhs[free_dofs_flat]
        diag_mat = diag_mat[free_dofs_flat]

    # Solve while assuming a diagonal tangent
    sol = jnp.multiply((1 / diag_mat), rhs)

    # Compose solution dofs
    if nodal_imposition:
        if isinstance(settings["dirichlet dofs"], (dict, FrozenDict)):
            free_dofs_dict = {
                key: jnp.invert(jnp.asarray(val))
                for key, val in settings["dirichlet dofs"].items()
            }
            sol = utility.reshape_as(sol, dofs)
            return utility.mask_op(dofs, free_dofs_dict, sol)
        else:
            return utility.mask_op(dofs, free_dofs_flat, sol)
    else:
        return utility.reshape_as(sol, dofs)

@utility.jit_with_docstring(static_argnames=["static_settings", "**kwargs"])
def solve_newton(
    dofs, settings, static_settings, newton_tol=1e-8, maxiter=30, **kwargs
    ):
    """
    Solves a nonlinear system using the Newton-Raphson method.

    This function is a wrapper for the damped Newton method with a damping coefficient of 1.0,
    effectively performing standard Newton-Raphson iterations.

    Args:
      dofs (jnp.ndarray or dict): Degrees of freedom or initial guess for the solution.
      settings (dict): Dictionary containing various settings and parameters required for assembling the problem.
      static_settings (dict): Dictionary containing static settings such as solution structure and solver backend.
      newton_tol (float, optional): Tolerance for the Newton method convergence criterion. Default is 1e-8.
      maxiter (int, optional): Maximum number of iterations for the Newton method. Default is 30.
      **kwargs (dict): Additional keyword arguments passed to the solver function.

    Returns:
      tuple: A tuple containing the following elements:
        - sol (jnp.ndarray): The solution obtained by solving the nonlinear system using the Newton method.
        - infos (tuple): Additional information about the solution process, including:
            - num_iterations (int): The number of iterations performed.
            - residual_norm (float): The norm of the residual at the solution.
            - diverged (bool): Flag indicating whether the method diverged.
    """
    return solve_damped_newton(
        dofs, settings, static_settings, newton_tol, 1.0, maxiter
    )

@utility.jit_with_docstring(static_argnames=["static_settings", "**kwargs"])
def solve_damped_newton(
    dofs,
    settings,
    static_settings,
    newton_tol=1e-8,
    damping_coefficient=0.8,
    maxiter=30,
    **kwargs,
    ):
    """
    Solves a nonlinear system using the damped Newton method.

    This function performs damped Newton iterations to solve a nonlinear system,
    with support for nodal imposition of Dirichlet boundary conditions specified as a boolean tuple-tree in static_settings['dirichlet dofs'].

    In this function the information needed for solver.damped_newton is prepared and the function is then called.

    Args:
      dofs (jnp.ndarray or dict): Degrees of freedom or initial guess for the solution.
      settings (dict): Dictionary containing various settings and parameters required for assembling the problem.
      static_settings (dict): Dictionary containing static settings such as solution structure and solver backend.
      newton_tol (float, optional): Tolerance for the Newton method convergence criterion. Default is 1e-8.
      damping_coefficient (float, optional): Damping coefficient for the Newton updates. Default is 0.8.
      maxiter (int, optional): Maximum number of iterations for the Newton method. Default is 30.
      **kwargs (dict): Additional keyword arguments passed to the solver function.

    Returns:
      tuple: A tuple containing the following elements:
        - sol (jnp.ndarray): The solution obtained by solving the nonlinear system using the damped Newton method.
        - infos (tuple): Additional information about the solution process, including:
            - num_iterations (int): The number of iterations performed.
            - residual_norm (float): The norm of the residual at the solution.
            - diverged (bool): Flag indicating whether the method diverged.
    """
    residual_fun = lambda x_i: assembler.assemble_residual(
        x_i, settings, static_settings
    )
    lin_solve_fun = lambda x_i: solve_linear(x_i, settings, static_settings, **kwargs)

    nodal_imposition = "nodal imposition" in static_settings["solution structure"]
    if nodal_imposition:
        # Impose Dirichlet boundaries. Dirichlet dofs has to be concrete, therefore it is passed in static_settings as tuple of tuples
        if isinstance(settings["dirichlet dofs"], (dict, FrozenDict)):
            free_dofs_flat = {
                key: jnp.invert(jnp.asarray(val).flatten())
                for key, val in settings["dirichlet dofs"].items()
            }
        else:
            free_dofs_flat = jnp.invert(
                jnp.asarray(settings["dirichlet dofs"]).flatten()
            )
    else:
        free_dofs_flat = jnp.ones(utility.dict_flatten(dofs).shape, dtype=jnp.bool_)

    verbose = static_settings["verbose"]
    return damped_newton(
        dofs,
        residual_fun,
        lin_solve_fun,
        free_dofs_flat,
        newton_tol,
        maxiter,
        damping_coefficient,
        verbose=verbose,
    )

def damped_newton(
    dofs_0,
    residual_fun,
    lin_solve_fun,
    free_dofs,
    newton_tol,
    maxiter,
    damping_coefficient,
    verbose=1,
    ):
    """
    Performs damped Newton iterations to solve a nonlinear system.

    This function implements the damped Newton method for solving a nonlinear system. It updates
    the solution iteratively based on the residual and chosen linear solver for the Newton step,
    with a damping coefficient to control the step size.

    Args:
      dofs_0 (jnp.ndarray or dict): Initial guess for the degrees of freedom.
      residual_fun (callable): Function to compute the residual of the system.
      lin_solve_fun (callable): Function to solve the linearized system for the Newton step.
      free_dofs (jnp.ndarray or None): Boolean array indicating the free degrees of freedom. 
            If None, all degrees of freedom are free.      
      newton_tol (float): Tolerance for the Newton method convergence criterion.
      maxiter (int): Maximum number of iterations for the Newton method.
      damping_coefficient (float): Damping coefficient for the Newton updates.

    Returns:
      tuple: A tuple containing the following elements:
        - sol (jnp.ndarray): The solution obtained by solving the nonlinear system using the damped Newton method.
        - infos (tuple): Additional information about the solution process, including:
            - num_iterations (int): The number of iterations performed.
            - residual_norm (float): The norm of the residual at the solution.
            - diverged (bool): Flag indicating whether the method diverged.
    """
    def step(carry):
        dofs_i, itt, _, res_norm_old, _ = carry

        # Update formula of newton scheme
        delta_x_i = lin_solve_fun(dofs_i)

        # If free_dofs is None, apply update to all dofs, otherwise apply masking
        if free_dofs is not None:
            # Apply damping to the Newton step on free dofs
            delta_x_i = utility.mask_op(
                delta_x_i, free_dofs, mode="apply", ufunc=lambda x: damping_coefficient * x
            )
            # Update free dofs with the damped step
            dofs_i = utility.mask_op(dofs_i, free_dofs, delta_x_i, "add")

            # Set boundary conditions for dirichlet_dofs
            if isinstance(free_dofs, dict):
                dirichlet_dofs = {key: jnp.invert(val) for (key, val) in free_dofs.items()}
            else:
                dirichlet_dofs = jnp.invert(free_dofs)
            dofs_i = utility.mask_op(dofs_i, dirichlet_dofs, delta_x_i, "set")
        else:
            # If free_dofs is None, apply the update to all dofs directly
            dofs_i += damping_coefficient * delta_x_i

        # Compute residual for next step as convergence test
        residual = residual_fun(dofs_i)
        if free_dofs is not None:
            residual_flat = utility.dict_flatten(utility.mask_select(residual, free_dofs))
        else:
            residual_flat = utility.dict_flatten(residual)  # Use full residual if no mask
        res_norm = jnp.linalg.norm(residual_flat)
        not_stop = jnp.where(res_norm > newton_tol, True, False)

        def report():
            if verbose > 0:
                jax.debug.print(
                    "Residual after Newton iteration {x}: {y}", x=itt + 1, y=res_norm
                )
                if verbose > 1:
                    jax.debug.print("")

            # Check for divergence
            divergence = jnp.where(
                jnp.logical_and(res_norm / res_norm_old > 10, itt > 1), True, False
            )
            # If nan or inf in residual, set divergence
            divergence = jnp.where(
                jnp.any(jnp.logical_or(jnp.isnan(residual_flat), jnp.isinf(residual_flat))),
                True,
                divergence,
            )
            return jnp.invert(divergence), divergence

        def stop_newton():
            jax.debug.print("Warning: Newton scheme could not converge!")
            return False, True

        next_step, divergence = jax.lax.cond(itt < maxiter, report, stop_newton)
        return (
            dofs_i,
            itt + 1,
            jnp.logical_and(not_stop, next_step),
            res_norm,
            divergence,
        )

    def convergence_check(carry):
        _, _, not_stop, _, _ = carry
        return not_stop

    # Start Newton iteration loop
    sol, load_steps, _, res_norm, divergence = lax.while_loop(
        convergence_check, step, (dofs_0, 0, True, 0.0, False)
    )

    return (sol, (load_steps, res_norm, divergence))

### Linear solvers for different backends
@utility.jit_with_docstring(static_argnames=["static_settings", "**kwargs"])
def linear_solve_jax(dofs, settings, static_settings, **kwargs):
    """
    Solves a linear system of equations using JAX's matrix free solvers.

    This function performs a linear solve using different itterative solvers,
    optionally imposing Dirichlet boundary conditions and preconditioning and
    using different matrix or Hessian vector product (HVP) methods.

    Args:
      dofs (jnp.ndarray): Initial degrees of freedom.
      settings (dict): Dictionary of problem settings.
      static_settings (dict): Dictionary of static settings that do not change during iterations.
      **kwargs: Additional keyword arguments for the solver.

    Returns:
      jnp.ndarray: Solution degrees of freedom after solving the linear system.

    Hessian Vector Product (HVP) Types:
      - 'fwdrev': Forward and reverse mode differentiation.
      - 'revrev': Reverse mode differentiation (only for symmetric matrices).
      - 'assemble': Assembles the tangent matrix explicitly (not supported with nodal imposition, then call e.g. PetSc).
      - 'linearize': Uses JAX's linearize function.
      - Default: Uses 'fwdrev' if not specified or if an unsupported type is provided.

    Solver Types:
      - 'cg': Conjugate Gradient.
      - 'normal cg': Normal Conjugate Gradient.
      - 'gmres': Generalized Minimal Residual Method.
      - 'bicgstab': BiConjugate Gradient Stabilized.
      - 'lu': LU Decomposition.
      - 'cholesky': Cholesky Decomposition (converts tangent to dense mode; not supported with nodal imposition).
      - 'qr': QR Decomposition (not supported with nodal imposition).
      - 'jacobi': Jacobi Method.
      - Default: Uses 'bicgstab' if not specified or if an unsupported type is provided.

    Notes:
      - Dirichlet boundary conditions are imposed if 'nodal imposition' is specified as the solution structure.
      - The itterative solver can be preconditioned with 'jacobi'.
      - When using 'assemble' HVP type, the function will explicitly assemble the tangent matrix.
    """
    assert not isinstance(
        dofs, dict
    ), "dofs as dict are currently not implemented in linear_solve_jax."

    solver = static_settings["solver"]
    hvp_type = static_settings["hvp type"]
    nodal_imposition = "nodal imposition" in static_settings["solution structure"]

    if nodal_imposition:
        # Impose Dirichlet boundaries. Dirichlet dofs has to be concrete, therefore it is passed in static_settings as tuple of tuples
        dirichlet_dofs_flat = jnp.asarray(settings["dirichlet dofs"]).flatten()
        dirichlet_conditions = settings["dirichlet conditions"].flatten()
        free_dofs_flat = jnp.invert(dirichlet_dofs_flat)

        # Impose nodal dofs
        flat_dofs = dofs.flatten()
        idx = jnp.arange(flat_dofs.shape[0])[dirichlet_dofs_flat]
        flat_dofs = flat_dofs.at[idx].set(dirichlet_conditions[idx])
        dofs = flat_dofs.reshape(dofs.shape)
        free_dofs = dofs.flatten()[free_dofs_flat]

        def residual_fun(x):
            flat_dofs = dofs.flatten()
            idx = jnp.arange(flat_dofs.shape[0])[free_dofs_flat]
            flat_dofs = flat_dofs.at[idx].set(x)
            return assembler.assemble_residual(
                flat_dofs.reshape(dofs.shape), settings, static_settings
            ).flatten()[free_dofs_flat]

        def diag_assemble_fun(x):
            flat_dofs = dofs.flatten()
            idx = jnp.arange(flat_dofs.shape[0])[free_dofs_flat]
            flat_dofs = flat_dofs.at[idx].set(x)
            return assembler.assemble_tangent_diagonal(
                flat_dofs.reshape(dofs.shape), settings, static_settings
            ).flatten()[free_dofs_flat]

        rhs = -residual_fun(dofs.flatten()[free_dofs_flat])

    else:
        free_dofs = dofs
        residual_fun = lambda x: assembler.assemble_residual(
            x, settings, static_settings
        )
        mat_assemble_fun = lambda x: assembler.assemble_tangent(
            x, settings, static_settings
        )
        diag_assemble_fun = lambda x: assembler.assemble_tangent_diagonal(
            x, settings, static_settings
        )
        rhs = -residual_fun(dofs)

    # Type of Hessian vector product
    match hvp_type:
        case "fwdrev":

            def hvp_fwdrev(v):
                return jax.jvp(residual_fun, (free_dofs,), (v,))[1]

            hvp = hvp_fwdrev
        case "revrev":  # only works for symmetric matrices

            def hvp_revrev(v):
                return jax.jacrev(lambda x: jnp.vdot(residual_fun(x), v))(free_dofs)

            hvp = hvp_revrev
        case "assemble":
            assert (
                not nodal_imposition
            ), "hvp type 'assemble' is currently not supported for nodal imposition of DOFs."

            tangent = mat_assemble_fun(free_dofs)

            def hvp_assembled(v):
                mapped = tangent @ v.flatten()
                return jnp.reshape(mapped, v.shape)

            hvp = hvp_assembled
        case "linearize":

            def hvp_linearized(v):
                (_, linearized) = jax.linearize(residual_fun, free_dofs)
                return linearized(v)

            hvp = hvp_linearized
        case _:
            hvp = hvp_fwdrev
            print(
                "Type of hessian vector product has not been set or is not available. Using static_settings['matrix-free'] = 'revrev' as default."
            )

    # Preconditioning for itterative solvers
    if (
        solver == "cg"
        or solver == "bicgstab"
        or solver == "normal cg"
        or solver == "gmres"
    ):
        try:
            precond_type = static_settings["type of preconditioner"]
            match precond_type:
                case "jacobi":
                    # Inversion of diagonal part of tangent matrix as preconditioner
                    M = 1 / diag_assemble_fun(free_dofs)

                    def preconditioner(v):
                        preconditioned = jnp.multiply(M, v.flatten())
                        return jnp.reshape(preconditioned, free_dofs.shape)

                case _:

                    def preconditioner(v):
                        return v

                    print(
                        "Wrong preconditioner keyword. Continue without preconditioner."
                    )
        except KeyError:
            preconditioner = None
            pass

    # Select linear solver (itterative or direct)
    match solver:
        case "cg":
            (sol, _) = jax.scipy.sparse.linalg.cg(hvp, rhs, M=preconditioner, **kwargs)
        case "normal cg":
            sol = jaxopt.linear_solve.solve_normal_cg(hvp, rhs, **kwargs)
        case "gmres":
            (sol, _) = jax.scipy.sparse.linalg.gmres(
                hvp, rhs, M=preconditioner, **kwargs
            )
        case "bicgstab":
            (sol, _) = jax.scipy.sparse.linalg.bicgstab(
                hvp, rhs, M=preconditioner, **kwargs
            )
        case "lu":
            sol = jaxopt.linear_solve.solve_lu(hvp, rhs)
        case "cholesky":
            assert (
                not nodal_imposition
            ), "solver 'cholesky' is currently not supported for nodal imposition of DOFs."
            chol, lower = jax.scipy.linalg.cho_factor(mat_assemble_fun(dofs).todense())
            sol = jnp.reshape(
                jax.scipy.linalg.cho_solve((chol, lower), rhs.flatten()), dofs.shape
            )
        case "qr":
            assert (
                not nodal_imposition
            ), "solver 'qr' is currently not supported for nodal imposition of DOFs."
            bcoo_tangent = mat_assemble_fun(dofs)
            bcsr_tangent = sparse.BCSR.from_bcoo(bcoo_tangent).sum_duplicates(
                nse=bcoo_tangent.nse
            )
            sol = sparse.linalg.spsolve(
                bcsr_tangent.data,
                bcsr_tangent.indices,
                bcsr_tangent.indptr,
                rhs.flatten(),
            ).reshape(dofs.shape)
        case "jacobi":
            diag = diag_assemble_fun(free_dofs)
            sol = jacobi_method(hvp, diag, free_dofs, rhs, **kwargs)
        case _:
            (sol, _) = jax.scipy.sparse.linalg.bicgstab(
                hvp, rhs, M=preconditioner, **kwargs
            )
            print(
                "Solver name not set or not available in combination with this solver type. Using static_settings['solver'] = 'bicgstab' as default."
            )

    if static_settings["verbose"] >= 1:
        residual = rhs - hvp(sol)
        jax.debug.print(
            "The relative residual is: {value}",
            value=jnp.linalg.norm(residual) / jnp.linalg.norm(rhs),
        )

    # Compose solution dofs
    if nodal_imposition:
        flat_dofs = dofs.flatten()
        flat_dofs = flat_dofs.at[idx].set(dirichlet_conditions[idx])
        idx = jnp.arange(flat_dofs.shape[0])[free_dofs_flat]
        flat_dofs = flat_dofs.at[idx].set(sol)
        dofs = flat_dofs.reshape(dofs.shape)
        return dofs
    else:
        return sol

def scipy_assembling(tangent_with_duplicates, verbose, free_dofs):
    """
    Convert a JAX BCOO matrix to a SciPy CSR matrix while summing duplicates.

    This function converts a JAX BCOO matrix, which may contain duplicate entries,
    into a SciPy CSR matrix. It optionally deletes rows and columns corresponding
    to Dirichlet degrees of freedom.

    Args:
      tangent_with_duplicates (jax.experimental.sparse.BCOO): The input JAX BCOO matrix.
      verbose (int): Verbosity level. If >= 2, timing information is printed.
      free_dofs (array or None): Boolean array indicating which degrees of freedom are free.
                                  If not None, rows and columns corresponding to Dirichlet DOFs
                                  are removed from the matrix.

    Returns:
      scipy.sparse.csr_matrix: The converted and possibly reduced CSR matrix.
    """
    if verbose >= 2:
        start = time.time()

    data = tangent_with_duplicates.data
    indices = tangent_with_duplicates.indices
    rows = indices[:, 0]
    cols = indices[:, 1]

    # This is currently done on CPU and seems to be one of the computational bottlenecks when using GPU
    tangent_coo = scp.sparse.coo_matrix(
        (data, (rows, cols)), shape=tangent_with_duplicates.shape
    )

    tangent_csr = scp.sparse.csr_matrix(tangent_coo)

    # Row deletion for Dirichlet-DOFs
    if free_dofs is not None:
        # Deleting rows and columns
        tangent_csr = tangent_csr[:, free_dofs]
        tangent_csr = tangent_csr[free_dofs]

    if verbose >= 2:
        print("Time for summing duplicates: ", time.time() - start)

    return tangent_csr

def linear_solve_petsc(mat, rhs, n_fields, solver, pc_type, verbose, free_dofs, tol=1e-8, **kwargs):
    """
    Solve a linear system using the PETSc solver (requires PETSc and petsc4py to be installed).

    This function solves a linear system using PETSc, with options for different
    solvers and preconditioners. The input matrix is first converted to a SciPy
    CSR matrix, and rows and columns corresponding to Dirichlet DOFs are optionally
    removed.

    Args:
      mat (jax.experimental.sparse.BCOO): The input matrix in JAX BCOO format.
      rhs (jnp.ndarray): The (reduced/free) right-hand side vector.
      n_fields (int): The number of fields.
      solver (str): The type of solver to use.
      pc_type (str): The type of preconditioner to use.
      verbose (int): Verbosity level. If >= 1, timing and solver information is printed.
      free_dofs (array or None): Boolean array indicating which degrees of freedom are free.
      tol (float): The relative tolerance for the solver.
      **kwargs: Additional keyword arguments for the solver.

    Returns:
      jax.numpy.array: The solution vector.

    Notes:
      - The solver settings can also be set from the command line. See PETSc and petsc4py documentation.
    """
    try:
        import petsc4py

        petsc4py.init(sys.argv)
        from petsc4py import PETSc
    except ModuleNotFoundError:
        print("Linear solver requires petsc and petsc4py")

    if free_dofs is not None:
        reduced_rhs = rhs[free_dofs]
        n_dofs = reduced_rhs.shape[0]
    else:
        reduced_rhs = rhs
        n_dofs = rhs.shape[0]

    # Transform matrix to csr format and sum duplicates
    tangent_csr = scipy_assembling(mat, verbose, free_dofs)
    if verbose >= 2:
        start = time.time()

    # Load to petsc
    mat_petsc = PETSc.Mat().createAIJ(
        size=tangent_csr.shape,
        csr=(
            tangent_csr.indptr.astype(PETSc.IntType),
            tangent_csr.indices.astype(PETSc.IntType),
            tangent_csr.data,
        ),
    )
    mat_petsc.setFromOptions()
    mat_petsc.setBlockSize(n_fields)

    if verbose >= 2:
        print("To PETSc transformation time: ", time.time() - start)
        print("Matrix infos: ")
        print()
        print(mat_petsc.getInfo())
        start = time.time()

    # Initialization of right hand side and solution vector
    b = PETSc.Vec().createSeq(n_dofs)
    b.setFromOptions()
    b.setArray(np.array(reduced_rhs))
    x = PETSc.Vec().createSeq(n_dofs)
    x.setFromOptions()

    # Solver settings
    rtol = tol
    ksp = PETSc.KSP().create()
    ksp.setTolerances(rtol=rtol, **kwargs)
    ksp.setOperators(mat_petsc)
    ksp.setType(solver)
    ksp.setFromOptions()
    ksp.setConvergenceHistory()
    ksp.getPC().setType(pc_type)
    ksp.getPC().setFromOptions()

    # Monitoring
    if verbose >= 2:
        print("Iteration   Residual")

        def monitor(ksp, its, rnorm):
            print("%5d      %20.15g" % (its, rnorm))

        ksp.setMonitor(monitor)

    # Solving
    ksp.solve(b, x)

    # Fill solution vector with computed values
    if free_dofs is not None:
        sol = jnp.zeros(rhs.shape)
        sol = sol.at[free_dofs].set(x.getArray())
    else:
        sol = jnp.array(x.getArray())

    if verbose >= 2:
        residual = mat_petsc * x - b
        print("Itterative linear solver time: ", time.time() - start)
        print("Number of iterations: ", ksp.getIterationNumber())
        print("Type: ", ksp.getType())
        print("Tolerances: ", ksp.getTolerances())
        print(f"The relative residual is: {residual.norm() / (b.norm() + 1e-12)}.")
    return sol

def linear_solve_pardiso(mat, rhs, solver, verbose, free_dofs):
    """
    Solve a linear system using the PARDISO solver (requires Intel MKL and pypardiso('lu') or sparse_dot_mkl('qr') to be installed).

    This function solves a linear system using PARDISO, with options for different
    solver types. The input matrix is first converted to a SciPy CSR matrix, and
    rows and columns corresponding to Dirichlet DOFs are optionally removed.

    Args:
      mat (jax.experimental.sparse.BCOO): The input matrix in JAX BCOO format.
      rhs (jnp.ndarray): The (reduced/free) right-hand side vector.
      solver (str): The type of solver to use ('lu' or 'qr').
      verbose (int): Verbosity level. If >= 2, timing information is printed.
      free_dofs (array or None): Boolean array indicating which degrees of freedom are free.

    Returns:
      jax.numpy.array: The solution vector.
    """
    # Transform matrix to csr format and sum duplicates
    tangent_csr = scipy_assembling(mat, verbose, free_dofs)
    if verbose >= 2:
        start = time.time()

    # Prepare right hand side
    if free_dofs is not None:
        b = np.asarray(rhs[free_dofs])
    else:
        b = np.asarray(rhs)

    if solver == "lu":
        try:
            import pypardiso
        except ModuleNotFoundError:
            print("Linear solver requires the installation of pypardiso.")

        # ToDo: make use of symmetries and other settings available#, msglvl=verbose, iparm=iparm
        # pypardiso_solver = pypardiso.PyPardisoSolver(mtype=11) # spd: 2
        # x = pypardiso.spsolve(tangent_csr, b, solver=pypardiso_solver)
        x = pypardiso.spsolve(tangent_csr, b)
    elif solver == "qr":
        try:
            import sparse_dot_mkl
        except ModuleNotFoundError:
            print("Linear solver requires the installation of sparse_dot_mkl.")

        x = sparse_dot_mkl.sparse_qr_solve_mkl(tangent_csr, b)
    else:
        assert False, "Type of solver not supported. Choose 'lu' or 'qr'"

    # Fill solution vector with computed values
    if free_dofs is not None:
        sol = jnp.zeros(rhs.shape)
        sol = sol.at[free_dofs].set(x)
    else:
        sol = jnp.array(x)

    if verbose >= 2:
        residual = b - tangent_csr * x
        print(
            f"The relative residual after linear solve is: {np.linalg.norm(residual) / (np.linalg.norm(b) + 1e-12)}."
        )
        print("Linear solver time: ", time.time() - start)
    return sol

def linear_solve_pyamg(mat, rhs, solver, pc_type, verbose, free_dofs, **kwargs):
    """
    Solve a linear system using the PyAMG solver (requires pyamg to be installed).

    This function solves a linear system using PyAMG, an algebraic multi-grid solver with options
    for different solvers and preconditioners. The input matrix is first converted to a SciPy
    CSR matrix, and rows and columns corresponding to Dirichlet DOFs are optionally
    removed.

    Args:
      mat (jax.experimental.sparse.BCOO): The input matrix in JAX BCOO format.
      rhs (jnp.ndarray): The (reduced/free) right-hand side vector.
      solver (str): The type of solver to use ('cg', 'bcgs', or 'gmres').
      pc_type (str): The type of preconditioner to use ('ruge stuben' or 'smoothed aggregation').
      verbose (int): Verbosity level. If >= 1, timing and solver information is printed.
      free_dofs (array or None): Boolean array indicating which degrees of freedom are free.
      **kwargs: Additional keyword arguments for the solver.

    Returns:
      jax.numpy.array: The solution vector.
    """
    # Transform matrix to csr format and sum duplicates
    pyamg_tangent = scipy_assembling(mat, verbose, free_dofs)

    if verbose >= 2:
        start = time.time()

    # Set up solver
    try:
        import pyamg
    except ModuleNotFoundError:
        print("Linear solver requires the installation of pyamg.")

    if pc_type == "ruge stuben":
        ml = pyamg.ruge_stuben_solver(A=pyamg_tangent)
    elif pc_type == "smoothed aggregation":
        ml = pyamg.smoothed_aggregation_solver(A=pyamg_tangent)
    elif pc_type == "root node":
        ml = pyamg.rootnode_solver(A=pyamg_tangent)
    elif pc_type == "pairwise":
        ml = pyamg.pairwise_solver(A=pyamg_tangent)
    else:
        assert (
            False
        ), "Type of preconditioner not supported. Choose 'ruge stuben' or 'smoothed aggregation', 'root node' or 'pairwise'"

    if verbose >= 2:
        print(ml)
        print("Time for setting up multigrid preconditioner: ", time.time() - start)
        start = time.time()

    # Prepare right hand side
    if free_dofs is not None:
        b = np.asarray(rhs[free_dofs])
    else:
        b = np.asarray(rhs)

    residuals = []

    # Solving
    if solver == None:
        x = ml.solve(b, tol=1e-8, residuals=residuals, cycle='W')
    elif solver == "cg":
        x = ml.solve(b, accel=scp.sparse.linalg.cg, tol=1e-8, residuals=residuals, cycle='W')
    elif solver == "bcgs":
        x = ml.solve(b, accel=scp.sparse.linalg.bicgstab, tol=1e-8, residuals=residuals, cycle='W')
    elif solver == "gmres":
        x = ml.solve(b, accel=scp.sparse.linalg.gmres, tol=1e-8, residuals=residuals, cycle='W')
    else:
        assert False, "Type of solver not supported. Choose None, 'cg', 'bcgs' or 'gmres'"

    # Fill solution vector with computed values
    if free_dofs is not None:
        sol = jnp.zeros(rhs.shape)
        sol = sol.at[free_dofs].set(x)
    else:
        sol = jnp.array(x)

    if verbose >= 2:
        residual = b - pyamg_tangent * x
        print(
            f"The relative residual is: {np.linalg.norm(residual) / (np.linalg.norm(b) + 1e-12)}."
        )
        print("Itterative linear solver time: ", time.time() - start)

    if verbose >= 3:
        import matplotlib.pyplot as plt
        plt.semilogy(residuals/residuals[0], 'o-')
        plt.xlabel('iterations')
        plt.ylabel('normalized residual')
        plt.show()

    return sol

def linear_solve_scipy(mat, rhs, solver, verbose, free_dofs):
    """
    Solves a linear system using a specified SciPy solver.

    Args:
      mat (bcoo): JAX BCOO matrix representing the system's tangent matrix.
      rhs (jnp.ndarray): The (reduced/free) right-hand side vector.
      solver (str): Type of solver to use. Options are 'lapack' or 'umfpack'.
      verbose (int): Verbosity level for logging.
      free_dofs (jnp.ndarray): Boolean array indicating free degrees of freedom for Dirichlet boundary conditions.

    Returns:
      sol (jnp.ndarray): Solution vector to the linear system.
    """
    # Transform matrix to csr format and sum duplicates
    tangent_csr = scipy_assembling(mat, verbose, free_dofs)

    if verbose >= 2:
        start = time.time()

    # Prepare right hand side
    if free_dofs is not None:
        b = rhs[free_dofs]
    else:
        b = rhs

    if solver == "lapack":
        x = scp.sparse.linalg.spsolve(tangent_csr, b)
    elif solver == "umfpack":
        x = scp.sparse.linalg.spsolve(tangent_csr, b, use_umfpack=True)
    else:
        assert False, "Type of solver not supported. Choose 'lapack' or 'umfpack'"

    # Fill solution vector with computed values
    if free_dofs is not None:
        sol = jnp.zeros(rhs.shape)
        sol = sol.at[free_dofs].set(x)
    else:
        sol = jnp.array(x)

    if verbose >= 2:
        residual = b - tangent_csr * x
        print(
            f"The relative residual is: {np.linalg.norm(residual) / (np.linalg.norm(b) + 1e-12)}."
        )
        print("Direct solver time: ", time.time() - start)
    return sol

### Iterative solvers/smoothers
def jacobi_method(hvp_fun, diag, x_0, rhs, tol=1e-6, atol=1e-6, maxiter=1000):
    """
    Solve Ax = b using Jacobi iterations (experimental).

    Args:
      hvp_fun (function): Hessian vector product function.
      diag (jnp.ndarray): Diagonal of the Hessian matrix.
      x_0 (jnp.ndarray): Initial guess for the solution.
      rhs (jnp.ndarray): Right-hand side vector of the linear system.
      tol (float): Relative tolerance for convergence.
      atol (float): Absolute tolerance for convergence.
      maxiter (int): Maximum number of iterations.

    Returns:
      array: Solution vector x.
    """
    assert not isinstance(
        x_0, dict
    ), "dofs as dict are currently not implemented in jacobi_method."

    # Initialization
    inverse_diag = 1 / diag
    scaled_rhs = jnp.multiply(inverse_diag, rhs)
    rhs_squared = jnp.vdot(rhs, rhs)

    # Itterations
    def body_fun(value):
        x_k, k = value
        dx_k1 = scaled_rhs - jnp.multiply(inverse_diag, hvp_fun(x_k))
        x_k1 = x_k + dx_k1
        return (x_k1, k + 1)

    def cond_fun(value):
        x_k, k = value
        r_k = rhs - hvp_fun(x_k)
        r_k_squared = jnp.vdot(r_k, r_k)
        return k < maxiter  # and r_k_squared > jnp.max(tol**2 * rhs_squared, atol**2)

    x_final, *_ = lax.while_loop(cond_fun, body_fun, (x_0, 0))

    return x_final

def damped_jacobi_relaxation(hvp_fun, diag, x_0, rhs, damping_factor=0.3333333, **kwargs):
    """
    Damped Jacobi smoother (experimental).

    Args:
      hvp_fun (callable): Hessian vector product function.
      diag (jnp.ndarray): Diagonal of the Hessian matrix.
      x_0 (jnp.ndarray): Initial guess for the solution.
      rhs (jnp.ndarray): Right-hand side vector of the linear system.
      damping_factor (float): Damping factor for the iterations (<=0.5 guarantees a good smoother).
      **kwargs (dict): Additional keyword arguments for customization.

    Returns:
      array: Solution vector x.
    """
    assert not isinstance(
        x_0, dict
    ), "dofs as dict are currently not implemented in damped_jacobi_relaxation."

    # Initialization
    inverse_diag = 1 / diag
    scaled_rhs = jnp.multiply(inverse_diag, rhs)

    # Itterations
    def body_fun(x_k, idx):
        dx_k1 = scaled_rhs - jnp.multiply(inverse_diag, hvp_fun(x_k))
        x_k1 = x_k + damping_factor * dx_k1
        return x_k1, None

    iterations = 20
    x_final, *_ = lax.scan(body_fun, init=x_0, xs=None, length=iterations)
    return x_final



# assembler.py
# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

"""
Module for assembling and integrating functionals, residuals and tangents, supporting both 
'sparse' and 'dense' modes and 'user potentials', 'user residuals' and 'user elements'.

The assembly modes 'dense' and 'sparse' use JAX's automatic differentiation capabilities to 
determine the global residual and the tangent for the given values of the degrees of freedom 
and the models and variational schemes specified in static_settings.
In the 'dense' mode, the tangent matrix is returned as a jnp.ndarray, while in the 'sparse' 
mode, it is returned as a jax.experimental.BCOO matrix with duplicates. 
In the 'user potential', 'user residual' and 'user element' modes, the global residual and 
tangent can be assembled based on user-defined element-wise contributions. 
Depending on the execution location, the entries are calculated on CPU or GPU. 
For an efficient calculation, JAX's automatic vectorization transformation (vmap) is used.
The summation of duplicates is then carried out within the solver module. Currently, 
SciPy is used for this on the CPU.
"""

from functools import partial

import jax.numpy as jnp
import jax
from jax import vmap, jacrev, jacfwd, hessian, jvp, linearize, vjp, custom_jvp
from jax.tree import map as treemap
from jax.experimental import sparse

from autopdex import variational_schemes
from autopdex.utility import jit_with_docstring, dict_zeros_like, dict_flatten, reshape_as


# TODO: change vmaps to _batched_map in order to reduce memory consumption


## Helper functions
def _get_indices(connectivity, dofs):
    """
    Constructs the global indices for the assembly of the tangent matrix.

    Args:
        connectivity (array or dict): Connectivity array or dictionary of connectivity arrays.
        dofs (array or dict): DOFs array or dictionary of DOFs.

    Returns:
        indices (jnp.ndarray): Array of indices for the sparse matrix.
    """
    if callable(dofs):
        dofs = dofs(0.)

    if isinstance(dofs, dict):
        keys = dofs.keys()

        # Field-Offsets
        field_offsets = {}
        current_offset = 0
        for field in keys:
            field_offsets[field] = current_offset
            field_size = dofs[field].size
            current_offset += field_size

        indices_list = []

        num_elems = connectivity[next(iter(keys))].shape[0]
        elem_indices = jnp.arange(num_elems, dtype=int)

        for field_i in keys:
            for field_j in keys:

                def one_elem_indices(elem_idx):
                    # Global DOFs for field_i
                    conn_i = connectivity[field_i][elem_idx]
                    if dofs[field_i].ndim == 1:
                        dofs_per_node_i = 1
                    else:
                        dofs_per_node_i = dofs[field_i].shape[-1]
                    field_offset_i = field_offsets[field_i]
                    dof_local_i = jnp.arange(dofs_per_node_i, dtype=int)
                    dof_indices_i = (
                        field_offset_i + conn_i[:, None] * dofs_per_node_i + dof_local_i
                    )
                    global_dofs_i = jnp.asarray(dof_indices_i, dtype=int).flatten()

                    # Global DOFs for field_j
                    conn_j = connectivity[field_j][elem_idx]
                    if dofs[field_j].ndim == 1:
                        dofs_per_node_j = 1
                    else:
                        dofs_per_node_j = dofs[field_j].shape[-1]

                    field_offset_j = field_offsets[field_j]
                    dof_local_j = jnp.arange(dofs_per_node_j, dtype=int)
                    dof_indices_j = (
                        field_offset_j + conn_j[:, None] * dofs_per_node_j + dof_local_j
                    )
                    global_dofs_j = jnp.asarray(dof_indices_j, dtype=int).flatten()

                    # Generate indices
                    row_indices = jnp.repeat(global_dofs_i, global_dofs_j.size)
                    col_indices = jnp.tile(global_dofs_j, global_dofs_i.size)
                    indices = jnp.stack([row_indices, col_indices], axis=-1)
                    return indices

                # Vectorize over elements
                all_elem_indices = vmap(one_elem_indices)(elem_indices)
                indices = all_elem_indices.reshape(-1, 2)
                indices_list.append(indices)

        # Concatenate all indices
        indices = jnp.concatenate(indices_list, axis=0)
        return indices
    else:
        # dofs is array
        if dofs.ndim == 1:
            dofs_per_node = 1
        else:
            dofs_per_node = dofs.shape[-1]

        def one_elem_idx(neighb):
            global_dofs = neighb[:, None] * dofs_per_node + jnp.arange(dofs_per_node)
            global_dofs = global_dofs.flatten()
            n_dofs_element = global_dofs.size

            row_indices = jnp.repeat(global_dofs, n_dofs_element)
            col_indices = jnp.tile(global_dofs, n_dofs_element)
            indices = jnp.stack([row_indices, col_indices], axis=-1)
            return indices.astype(int)

        all_elem_indices = vmap(one_elem_idx)(connectivity)
        indices = all_elem_indices.reshape(-1, 2)
        return indices

def _get_element_quantities(dofs, settings, static_settings, set):
    """
    Extracts element-dependent quantities for the specified set.

    Args:
        dofs (jnp.ndarray or dict or callable): Degrees of freedom. Can be a function of time for transient problems.
        settings (dict): Settings dictionary.
        static_settings (dict or flax.core.FrozenDict): Static settings dictionary.
        set (int): The domain number.

    Returns:
        tuple: (model_fun, local_dofs, local_node_coor, elem_numbers, connectivity)
    """
    model_fun = static_settings["model"][set]
    x_nodes = settings["node coordinates"]
    dofs_is_fun = True if callable(dofs) else False
    if dofs_is_fun:
        dofs_is_dict = True if isinstance(dofs(0.), dict) else False
    else:
        dofs_is_dict = True if isinstance(dofs, dict) else False

    # Warning if it was defined in static_settings
    assert "connectivity" not in static_settings, \
        "'connectivity' has been moved to 'settings' in order to reduce compile time. \
        Further, you should not transform it to a tuple of tuples anymore."

    connectivity = settings["connectivity"][set]

    if dofs_is_dict:
        assert isinstance(
            x_nodes, dict
        ), "If 'dofs' is a dict, 'settings['node coordinates']' must also be a dict."
        assert isinstance(
            connectivity, dict
        ), "If 'dofs' is a dict, 'settings['connectivity'][set]' must also be a dict."

        elem_numbers = jnp.arange(connectivity[next(iter(connectivity))].shape[0])
    else:
        elem_numbers = jnp.arange(connectivity.shape[0])

    return model_fun, x_nodes, elem_numbers, connectivity

def _get_element_quantities_2(dofs, settings, static_settings, set):
    if callable(dofs):
        dofs = dofs(0.)
    assert isinstance(
        dofs, jnp.ndarray
    ), "Variational schemes do currently not support dofs as dicts."

    connectivity = settings["connectivity"][set]
    variational_scheme = static_settings["variational scheme"][set]
    x_int = settings["integration coordinates"][set]
    w_int = settings["integration weights"][set]
    int_point_numbers = jnp.arange(0, x_int.shape[0], 1)
    return connectivity, variational_scheme, x_int, w_int, int_point_numbers

def _extract_local_dofs_and_coor(dofs, node_list, x_nodes):
    # If DOFs are a function of time (for transient problems, forward them as a function of time)
    if callable(dofs):
        local_dofs = lambda t: treemap(lambda x, y: x.at[y].get(), dofs(t), node_list)
        # local_dofs = _make_elem_dofs_fun(dofs, node_list)
    else:
        local_dofs = treemap(lambda x, y: x.at[y].get(), dofs, node_list)
    local_node_coor = treemap(lambda x, y: x.at[y].get(), x_nodes, node_list)
    return local_dofs, local_node_coor

def _extract_local_dofs_and_coor_2(dofs, int_point_number, x_int, w_int, connectivity):
    x_i = x_int[int_point_number]
    w_i = w_int[int_point_number]
    if callable(dofs):
        local_dofs = lambda t: treemap(lambda x, y: x.at[y].get(), dofs(t), connectivity[int_point_number])
        # local_dofs = _make_elem_dofs_fun(dofs, connectivity[int_point_number])
    else:
        local_dofs = treemap(lambda x, y: x.at[y].get(), dofs, connectivity[int_point_number])
    return x_i, w_i, local_dofs

def _get_tangent_diagonal(tangent_contributions, connectivity, dofs):
    """
    Assembles the diagonal entries of the tangent matrix from the tangent contributions,
    connectivity, and degrees of freedom.

    Args:
        tangent_contributions (dict or jnp.ndarray): Tangent contributions from the model function.
        connectivity (dict or jnp.ndarray): Connectivity information for elements.
        dofs (dict or jnp.ndarray): Degrees of freedom.

    Returns:
        jnp.ndarray: The assembled diagonal of the tangent matrix.
    """
    if callable(dofs):
        dofs = dofs(0.)

    # Total number of DOFs
    num_dofs = (
        dofs.size if not isinstance(dofs, dict) else sum(v.size for v in dofs.values())
    )

    if isinstance(dofs, dict):
        keys = dofs.keys()
        n_elems = next(iter(connectivity.values())).shape[0]  # Number of elements

        # Compute field offsets for global DOF numbering
        field_offsets = {}
        current_offset = 0
        for key in keys:
            field_offsets[key] = current_offset
            field_size = dofs[key].size  # Total DOFs in the field
            current_offset += field_size

        # Initialize the global diagonal vector
        diag = jnp.zeros(num_dofs)

        # Iterate over fields to assemble diagonal contributions
        for key in keys:
            # Extract tangent contributions for field [key][key]
            tc = tangent_contributions[key][key]
            # tc has shape: (n_elems, nodes_per_element_i, dofs_per_node_i, nodes_per_element_j, dofs_per_node_j)

            # Reshape tc to (n_elems, element_dofs, element_dofs)
            conn = connectivity[key]  # Shape: (n_elems, nodes_per_element)
            nodes_per_element = conn.shape[1]
            if dofs[key].ndim == 1:
                dofs_per_node = 1
            else:
                dofs_per_node = dofs[key].shape[-1]
            element_dofs = nodes_per_element * dofs_per_node

            # Reshape tc
            tc = tc.reshape(n_elems, element_dofs, element_dofs)

            # Extract diagonal contributions
            diagonal_contributions = jnp.diagonal(
                tc, axis1=1, axis2=2
            )  # Shape: (n_elems, element_dofs)

            # Compute global DOF indices
            dof_local = jnp.arange(dofs_per_node)
            dof_indices = (
                field_offsets[key]
                + conn[:, :, None] * dofs_per_node
                + dof_local[None, None, :]
            )  # Shape: (n_elems, nodes_per_element, dofs_per_node)
            global_dofs = dof_indices.reshape(
                n_elems, element_dofs
            )  # Shape: (n_elems, element_dofs)

            # Flatten indices and values
            diag_indices = global_dofs.flatten().astype(int)
            diag_values = diagonal_contributions.flatten()

            # Sum into the global diagonal vector
            diag = diag.at[diag_indices].add(diag_values)

    else:
        # For the array case
        n_elems = connectivity.shape[0]
        conn = connectivity  # Shape: (n_elems, nodes_per_element)
        nodes_per_element = conn.shape[1]
        dofs_per_node = dofs.shape[-1]
        element_dofs = nodes_per_element * dofs_per_node

        # Reshape tangent_contributions
        tc = tangent_contributions.reshape(n_elems, element_dofs, element_dofs)

        # Extract diagonal contributions
        diagonal_contributions = jnp.diagonal(
            tc, axis1=1, axis2=2
        )  # Shape: (n_elems, element_dofs)

        # Compute global DOF indices
        dof_local = jnp.arange(dofs_per_node)
        dof_indices = (
            conn[:, :, None] * dofs_per_node + dof_local[None, None, :]
        )  # Shape: (n_elems, nodes_per_element, dofs_per_node)
        global_dofs = dof_indices.reshape(
            n_elems, element_dofs
        )  # Shape: (n_elems, element_dofs)

        # Flatten indices and values
        diag_indices = global_dofs.flatten().astype(int)
        diag_values = diagonal_contributions.flatten()

        # Sum into the global diagonal vector
        diag = jnp.zeros(num_dofs)
        diag = diag.at[diag_indices].add(diag_values)

    return diag

def _get_residual(residual_contributions, connectivity, dofs):
    """
    Assembles the global residual vector from the residual contributions,
    connectivity, and degrees of freedom, returning a residual with the same
    structure as dofs.

    Args:
        residual_contributions (dict or jnp.ndarray): Residual contributions from the model function.
        connectivity (dict or jnp.ndarray): Connectivity information for elements.
        dofs (dict or jnp.ndarray): Degrees of freedom.

    Returns:
        dict or jnp.ndarray: The assembled residual vector with the same structure as dofs.
    """
    if callable(dofs):
        dofs = dofs(0.)

    if isinstance(dofs, dict):
        keys = dofs.keys()
        n_elems = connectivity[next(iter(keys))].shape[0]
        
        # Initialize the residual dictionary
        residual = {}

        # Iterate over fields to assemble residual contributions
        for key in keys:
            # Extract residual contributions for field [key]
            rc = residual_contributions[
                key
            ]  # Shape: (n_elems, nodes_per_element, dofs_per_node)

            # Reshape rc to (n_elems, element_dofs)
            conn = connectivity[key]  # Shape: (n_elems, nodes_per_element)
            nodes_per_element = conn.shape[1]
            if dofs[key].ndim == 1:
                dofs_per_node = 1
            else:
                dofs_per_node = dofs[key].shape[-1]
            element_dofs = nodes_per_element * dofs_per_node

            # Reshape rc
            rc = rc.reshape(n_elems, element_dofs)  # Shape: (n_elems, element_dofs)

            # Compute global DOF indices
            dof_local = jnp.arange(dofs_per_node)
            dof_indices = (
                conn[:, :, None] * dofs_per_node + dof_local[None, None, :]
            )  # Shape: (n_elems, nodes_per_element, dofs_per_node)
            global_dofs = dof_indices.reshape(
                n_elems, element_dofs
            )  # Shape: (n_elems, element_dofs)

            # Flatten indices and values
            residual_indices = global_dofs.flatten().astype(int)
            residual_values = rc.flatten()

            # Initialize the residual array for this field
            field_residual = jnp.zeros_like(dofs[key]).flatten()

            # Sum into the field residual vector
            field_residual = field_residual.at[residual_indices].add(residual_values)

            # Reshape back to the original shape
            field_residual = field_residual.reshape(dofs[key].shape)

            # Assign to the residual dictionary
            residual[key] = field_residual

        return residual

    else:
        # For the array case
        n_elems = connectivity.shape[0]
        conn = connectivity  # Shape: (n_elems, nodes_per_element)
        nodes_per_element = conn.shape[1]
        dofs_per_node = dofs.shape[-1]
        element_dofs = nodes_per_element * dofs_per_node

        # Reshape residual_contributions
        rc = residual_contributions.reshape(
            n_elems, element_dofs
        )  # Shape: (n_elems, element_dofs)

        # Compute global DOF indices
        dof_local = jnp.arange(dofs_per_node)
        dof_indices = (
            conn[:, :, None] * dofs_per_node + dof_local[None, None, :]
        )  # Shape: (n_elems, nodes_per_element, dofs_per_node)
        global_dofs = dof_indices.reshape(
            n_elems, element_dofs
        )  # Shape: (n_elems, element_dofs)

        # Flatten indices and values
        residual_indices = global_dofs.flatten().astype(int)
        residual_values = rc.flatten()

        # Initialize the residual array
        residual = jnp.zeros_like(dofs).flatten()

        # Sum into the residual vector
        residual = residual.at[residual_indices].add(residual_values)

        # Reshape back to the original shape
        residual = residual.reshape(dofs.shape)

        return residual

def _make_elem_dofs_fun(dofs, elem):
    """This function takes the function `dofs(t)` and returns basically lambda t: dofs(t)[elem].
    
    The difference is, that when this function is used under jacrev, it will allocate less memory.
    Supports only as many derivatives as are defined via the custom_jvp decorators
    """
    # @custom_jvp
    # def elem_dofs_ttt_f(t):
    #     dofs_ttt_ = jacfwd(jacfwd(jacfwd(dofs)))(t)
    #     elem_dofs_ttt_ = treemap(lambda x, y: x.at[y].get(), dofs_ttt_, elem)
    #     return elem_dofs_ttt_
    # @elem_dofs_ttt_f.defjvp
    # def elem_dofs_ttt_jvp(primals, tangents):
    #     t, = primals
    #     t_dot, = tangents
    #     elem_dofs_tttt = jacfwd(jacfwd(jacfwd(jacfwd(dofs))))(t)[elem]
    #     return elem_dofs_ttt_f(t), treemap(lambda x: x * t_dot, elem_dofs_tttt)
    # @custom_jvp
    # def elem_dofs_tt_f(t):
    #     dofs_tt_ = jacfwd(jacfwd(dofs))(t)
    #     elem_dofs_tt_ = treemap(lambda x, y: x.at[y].get(), dofs_tt_, elem)
    #     return elem_dofs_tt_
    # @elem_dofs_tt_f.defjvp
    # def elem_dofs_tt_jvp(primals, tangents):
    #     t, = primals
    #     t_dot, = tangents
    #     elem_dofs_ttt = elem_dofs_ttt_f(t)
    #     return elem_dofs_tt_f(t), treemap(lambda x: x * t_dot, elem_dofs_ttt)
    # @custom_jvp
    # def elem_dofs_t_f(t):
    #     dofs_t_ = jacfwd(dofs)(t)
    #     elem_dofs_t_ = treemap(lambda x, y: x.at[y].get(), dofs_t_, elem)
    #     return elem_dofs_t_
    # @elem_dofs_t_f.defjvp
    # def elem_dofs_t_jvp(primals, tangents):
    #     t, = primals
    #     t_dot, = tangents
    #     elem_dofs_tt = elem_dofs_tt_f(t)
    #     return elem_dofs_t_f(t), treemap(lambda x: x * t_dot, elem_dofs_tt)
    # @custom_jvp
    # def elem_dofs_f(t):
    #     dofs_ = dofs(t)
    #     elem_dofs_ = treemap(lambda x, y: x.at[y].get(), dofs_, elem)
    #     return elem_dofs_
    # @elem_dofs_f.defjvp
    # def elem_dofs_jvp(primals, tangents):
    #     t, = primals
    #     t_dot, = tangents
    #     elem_dofs_t = elem_dofs_t_f(t)
    #     return elem_dofs_f(t), treemap(lambda x: x * t_dot, elem_dofs_t)
    # return elem_dofs_f


    # @partial(custom_jvp, nondiff_argnums=(1, 2))
    # def elem_dofs_ttt_f(t, dofs, elem):
    #     dofs_ttt_ = jacfwd(jacfwd(jacfwd(dofs)))(t)
    #     elem_dofs_ttt_ = treemap(lambda x, y: x.at[y].get(), dofs_ttt_, elem)
    #     return elem_dofs_ttt_
    # @elem_dofs_ttt_f.defjvp
    # def elem_dofs_ttt_jvp(dofs, elem, primals, tangents):
    #     t, = primals
    #     t_dot, = tangents
    #     elem_dofs_tttt = jacfwd(jacfwd(jacfwd(jacfwd(dofs))))(t)[elem]
    #     return elem_dofs_ttt_f(t, dofs, elem), treemap(lambda x: x * t_dot, elem_dofs_tttt)
    # @partial(custom_jvp, nondiff_argnums=(1, 2))
    # def elem_dofs_tt_f(t, dofs, elem):
    #     dofs_tt_ = jacfwd(jacfwd(dofs))(t)
    #     elem_dofs_tt_ = treemap(lambda x, y: x.at[y].get(), dofs_tt_, elem)
    #     return elem_dofs_tt_
    # @elem_dofs_tt_f.defjvp
    # def elem_dofs_tt_jvp(dofs, elem, primals, tangents):
    #     t, = primals
    #     t_dot, = tangents
    #     elem_dofs_ttt = elem_dofs_ttt_f(t, dofs, elem)
    #     return elem_dofs_tt_f(t, dofs, elem), treemap(lambda x: x * t_dot, elem_dofs_ttt)
    @partial(custom_jvp, nondiff_argnums=(1, 2))
    def elem_dofs_f(t, dofs, elem):
        dofs_ = dofs(t)
        elem_dofs_ = treemap(lambda x, y: x[y], dofs_, elem)
        return elem_dofs_
    @elem_dofs_f.defjvp
    def elem_dofs_jvp(dofs, elem, primals, tangents):
        t, = primals
        t_dot, = tangents

        # @partial(custom_jvp, nondiff_argnums=(1, 2))
        # def elem_dofs_t_f(t, dofs, elem):
        #     dofs_t_ = jacfwd(dofs)(t)
        #     elem_dofs_t_ = treemap(lambda x, y: x.at[y].get(), dofs_t_, elem)
        #     return elem_dofs_t_
        # @elem_dofs_t_f.defjvp
        # def elem_dofs_t_jvp(dofs, elem, primals, tangents):
        #     t, = primals
        #     t_dot, = tangents
        #     # elem_dofs_tt = elem_dofs_tt_f(t, dofs, elem)
        #     elem_dofs_tt = jacfwd(jacfwd(dofs))(t)[elem]
        #     return elem_dofs_t_f(t, dofs, elem), treemap(lambda x: x * t_dot, elem_dofs_tt)

        # elem_dofs_t = elem_dofs_t_f(t, dofs, elem)
        elem_dofs_t = treemap(lambda a, b: a[b], jacfwd(dofs)(t), elem)
        return elem_dofs_f(t, dofs, elem), treemap(lambda x: x * t_dot, elem_dofs_t)
    local_dofs_fun = lambda t: elem_dofs_f(t, dofs, elem)
    return local_dofs_fun

def _batched_map(fun, elem_numbers, connectivity):
    body_fun = lambda i: fun(elem_numbers[i], jax.tree.map(lambda x: x[i], connectivity))
    num_dofs_per_elem = jax.eval_shape(lambda i: dict_flatten(body_fun(i)), 0).shape[0]
    return jax.lax.map(body_fun, jnp.arange(elem_numbers.shape[0]), batch_size=int(64000/num_dofs_per_elem))


### General assembling functions
@jit_with_docstring(static_argnames=["static_settings"], possibly_static_argnames=['dofs'])
def integrate_functional(dofs, settings, static_settings):
    """
    Integrate functional as sum over set of domains.

    Args:
      dofs (jnp.ndarray or dict): Degrees of freedom.
      settings (dict): Settings dictionary.
      static_settings (flax.core.FrozenDict): Static settings as frozen dictionary.

    Returns:
      float: The integrated functional value for current dofs.
    """
    # Loop over all sets of integration points/ domains
    num_sets = len(static_settings["assembling mode"])
    integrated_functional = 0
    for set in range(num_sets):
        assembling_mode = static_settings["assembling mode"][set]

        if assembling_mode == "dense":
            integrated_functional += dense_integrate_functional(
                dofs, settings, static_settings, set
            )
        elif assembling_mode == "sparse":
            integrated_functional += sparse_integrate_functional(
                dofs, settings, static_settings, set
            )
        elif assembling_mode == "user potential":
            integrated_functional += user_potential_integrate_functional(
                dofs, settings, static_settings, set
            )
        else:
            assert (
                False
            ), "Assembling mode can be either 'sparse' or 'dense' in integrate_functional"

    return integrated_functional

@jit_with_docstring(static_argnames=["static_settings"], possibly_static_argnames=['dofs'])
def assemble_residual(dofs, settings, static_settings):
    """
    Assemble residuals over set of domains.

    Args:
      dofs (jnp.ndarray or dict or callable): Degrees of freedom. 
        Can be a function of time for transient problems in combination with user_residuals.
      settings (dict): Settings dictionary.
      static_settings (flax.core.FrozenDict): Static settings as frozen dictionary.

    Returns:
      jnp.ndarray: The assembled residual.
    """
    # Loop over all sets of integration points/ domains
    num_sets = len(static_settings["assembling mode"])

    if isinstance(dofs(0.), dict) if callable(dofs) else isinstance(dofs, dict):
        assert all([isinstance(settings['connectivity'][0], dict),
                    isinstance(settings['node coordinates'], dict)]), \
                    "If the DOFs are a dict, the connectivity, node coordinates, dirichlet dofs, and dirichlet conditions must also be dicts."
        
    if callable(dofs):
        integrated_residual = dict_zeros_like(dofs(0.))
    else:
        integrated_residual = dict_zeros_like(dofs)
    for set in range(num_sets):
        assembling_mode = static_settings["assembling mode"][set]

        if assembling_mode == "dense":
            add = dense_assemble_residual(dofs, settings, static_settings, set)
        elif assembling_mode == "sparse":
            add = sparse_assemble_residual(dofs, settings, static_settings, set)
        elif assembling_mode == "user potential":
            add = user_potential_assemble_residual(dofs, settings, static_settings, set)
        elif assembling_mode == "user residual":
            add = user_residual_assemble_residual(dofs, settings, static_settings, set)
        elif assembling_mode == "user element":
            add = user_element_assemble_residual(dofs, settings, static_settings, set)
        else:
            assert (
                False
            ), "Assembling mode can be either 'sparse', 'dense' or 'user element'"

        # Handle both cases dict and jnp.ndarray
        if isinstance(add, dict):
            integrated_residual = treemap(lambda x, y: x + y, integrated_residual, add)
        else:
            integrated_residual += add

    return integrated_residual

@jit_with_docstring(static_argnames=["static_settings"], possibly_static_argnames=['dofs'])
def assemble_tangent_diagonal(dofs, settings, static_settings):
    """
    Assemble the diagonal of the tangent matrix.

    Args:
      dofs (jnp.ndarray or dict): Degrees of freedom.
      settings (dict): Settings dictionary.
      static_settings (flax.core.FrozenDict): Static settings as frozen dictionary.

    Returns:
      jnp.ndarray: The diagonal of the assembled tangent matrix.
    """
    # Loop over all sets of integration points/ domains
    num_sets = len(static_settings["assembling mode"])
    if callable(dofs):
        tangent_diagonal = jnp.zeros_like(dict_flatten(dofs(0.)))
    else:
        tangent_diagonal = jnp.zeros_like(dict_flatten(dofs))
    for set in range(num_sets):
        assembling_mode = static_settings["assembling mode"][set]
        if assembling_mode == "sparse":
            tangent_diagonal += sparse_assemble_tangent_diagonal(
                dofs, settings, static_settings, set
            )
        elif assembling_mode == "user potential":
            tangent_diagonal += user_potential_assemble_tangent_diagonal(
                dofs, settings, static_settings, set
            )
        elif assembling_mode == "user residual":
            tangent_diagonal += user_residual_assemble_tangent_diagonal(
                dofs, settings, static_settings, set
            )
        elif assembling_mode == "user element":
            tangent_diagonal += user_element_assemble_tangent_diagonal(
                dofs, settings, static_settings, set
            )
        else:
            assert (
                False
            ), "Assembling mode for assembling tangent diagonal supports currently only 'sparse' and 'user element'"
    return tangent_diagonal

@jit_with_docstring(static_argnames=["static_settings"], possibly_static_argnames=['dofs'])
def assemble_tangent(dofs, settings, static_settings):
    """
    Assemble the full (possibly sparse) tangent matrix.

    Args:
      dofs (jnp.ndarray or dict): Degrees of freedom.
      settings (dict): Settings dictionary.
      static_settings (flax.core.FrozenDict): Static settings as frozen dictionary.

    Returns:
      jnp.ndarray or sparse matrix: The assembled tangent matrix.
    """
    # ToDo: add symmetric mode

    num_sets = len(static_settings["assembling mode"])
    one_dense = "dense" in static_settings["assembling mode"]

    if isinstance(dofs, dict):
        num_dofs = sum(v.size for v in dofs.values())
    else:
        num_dofs = dofs.size

    try:
        sparsity_pattern = static_settings["known sparsity pattern"]
    except KeyError:
        sparsity_pattern = "none"

    match sparsity_pattern:
        case "none":
            if one_dense:
                integrated_tangent = jnp.zeros((num_dofs, num_dofs))
            else:
                integrated_tangent = sparse.empty(
                    (num_dofs, num_dofs), dtype=float, index_dtype=jnp.int_
                )

            # Loop over all sets of integration points/ domains
            for set in range(num_sets):
                assembling_mode = static_settings["assembling mode"][set]

                if assembling_mode == "dense":
                    integrated_tangent += dense_assemble_tangent(
                        dofs, settings, static_settings, set
                    )
                else:
                    if assembling_mode == "sparse":
                        add = sparse_assemble_tangent(
                            dofs, settings, static_settings, set
                        )
                    elif assembling_mode == "user potential":
                        add = user_potential_assemble_tangent(
                            dofs, settings, static_settings, set
                        )
                    elif assembling_mode == "user residual":
                        add = user_residual_assemble_tangent(
                            dofs, settings, static_settings, set
                        )
                    elif assembling_mode == "user element":
                        add = user_element_assemble_tangent(
                            dofs, settings, static_settings, set
                        )
                    else:
                        assert (
                            False
                        ), "Assembling mode can be either 'sparse', 'dense' or 'user element'"

                    if one_dense:
                        integrated_tangent += add.todense()
                    else:
                        integrated_tangent += add
        case "diagonal":
            # # Compute the diagonal tangent with sparsejac (not the diagonal of a tangent that is not diagonal)
            # residual_fun = lambda flat_dofs: assemble_residual(flat_dofs.reshape(dofs.shape), settings, static_settings).flatten()
            # with jax.ensure_compile_time_eval():
            #   data_and_indices = (jnp.ones((num_dofs,)),
            #                       vmap(lambda i: jnp.asarray([i, i]))(jnp.arange(0, num_dofs)))
            #   mat_shape = (num_dofs,num_dofs)
            #   sparsity = sparse.BCOO(data_and_indices, shape=mat_shape)
            #   sparse_diag_fun = sparsejac.jacfwd(residual_fun, sparsity=sparsity)
            # diag = sparse_diag_fun(dofs.flatten())
            # return diag

            diag = dict_flatten(
                assemble_tangent_diagonal(dofs, settings, static_settings)
            )
            indices = vmap(lambda i: jnp.asarray([i, i]))(jnp.arange(0, num_dofs))
            data_and_indices = (diag, indices)
            matrix_shape = (num_dofs, num_dofs)
            diag_mat = sparse.BCOO(data_and_indices, shape=matrix_shape)
            return diag_mat

        case _:
            assert False, "'known sparsity pattern' mode is not implemented."

    return integrated_tangent


### Dense assembling
@jit_with_docstring(static_argnames=["static_settings", "set"], possibly_static_argnames=['dofs'])
def dense_integrate_functional(dofs, settings, static_settings, set):
    """
    Dense integration of functional of specified domain.

    Args:
      dofs (jnp.ndarray or dict): Degrees of freedom.
      settings (dict): Settings dictionary.
      static_settings (flax.core.FrozenDict): Static settings as frozen dictionary.
      set (int): The domain number.

    Returns:
      float: The integrated functional value.
    """
    x_int = settings["integration coordinates"][set]
    w_int = settings["integration weights"][set]
    int_point_numbers = jnp.arange(0, x_int.shape[0], 1)

    def at_int_point(int_point_number):
        x_i = x_int[int_point_number]
        w_i = w_int[int_point_number]
        return variational_schemes.functional_at_int_point(x_i, w_i, int_point_number, dofs, settings, static_settings, set)

    functional_at_int_point_vj = vmap(at_int_point, (0,))
    integrated_functional = functional_at_int_point_vj(int_point_numbers).sum()

    return integrated_functional

@jit_with_docstring(static_argnames=["static_settings", "set"], possibly_static_argnames=['dofs'])
def dense_assemble_residual(dofs, settings, static_settings, set):
    """
    Dense assembly of residual of specified domain.

    Args:
      dofs (jnp.ndarray or dict): Degrees of freedom.
      settings (dict): Settings dictionary.
      static_settings (flax.core.FrozenDict): Static settings as frozen dictionary.
      set (int): The domain number.

    Returns:
      jnp.ndarray: The assembled residual.
    """
    return jacrev(dense_integrate_functional)(dofs, settings, static_settings, set)

@jit_with_docstring(static_argnames=["static_settings", "set"], possibly_static_argnames=['dofs'])
def dense_assemble_tangent(dofs, settings, static_settings, set):
    """
    Dense assembly of tangent of specified domain.

    Args:
      dofs (jnp.ndarray or dict): Degrees of freedom.
      settings (dict): Settings dictionary.
      static_settings (flax.core.FrozenDict): Static settings as frozen dictionary.
      set (int): The domain number.

    Returns:
      jnp.ndarray: The assembled tangent matrix.
    """
    assert isinstance(
        dofs, jnp.ndarray
    ), "Dense mode of tangent assembly does currently not support dofs as dicts."

    size = dict_flatten(dofs).size
    tangent = hessian(dense_integrate_functional)(dofs, settings, static_settings, set)
    return dict_flatten(tangent).reshape((size, size))


### Sparse assembling
@jit_with_docstring(static_argnames=["static_settings", "set"], possibly_static_argnames=['dofs'])
def sparse_integrate_functional(dofs, settings, static_settings, set):
    """
    Sparse integration of functional of specified domain.

    Args:
      dofs (jnp.ndarray or dict): Degrees of freedom.
      settings (dict): Settings dictionary.
      static_settings (flax.core.FrozenDict): Static settings as frozen dictionary.
      set (int): The domain number.

    Returns:
      float: The integrated functional value.
    """
    connectivity, variational_scheme, x_int, w_int, int_point_numbers = _get_element_quantities_2(dofs, settings, static_settings, set)

    def func_at_int_pt(int_point_number):
        x_i, w_i, local_dofs = _extract_local_dofs_and_coor_2(dofs, int_point_number, x_int, w_int, connectivity)
        return variational_schemes.functional_at_int_point(
            x_i, w_i, int_point_number, local_dofs, settings, static_settings, set
        )

    functional_at_int_point_vj = vmap(func_at_int_pt, (0,))
    return functional_at_int_point_vj(int_point_numbers).sum()

@jit_with_docstring(static_argnames=["static_settings", "set"], possibly_static_argnames=['dofs'])
def sparse_assemble_residual(dofs, settings, static_settings, set):
    """
    Sparse assembly of residual of specified domain.

    Args:
      dofs (jnp.ndarray or dict): Degrees of freedom.
      settings (dict): Settings dictionary.
      static_settings (flax.core.FrozenDict): Static settings as frozen dictionary.
      set (int): The domain number.

    Returns:
      jnp.ndarray: The assembled residual.
    """
    connectivity, variational_scheme, x_int, w_int, int_point_numbers = _get_element_quantities_2(dofs, settings, static_settings, set)

    if (
        variational_scheme == "least square pde loss"
        or variational_scheme == "least square function approximation"
    ):
        def func_at_int_pt(int_point_number):
            x_i, w_i, local_dofs = _extract_local_dofs_and_coor_2(dofs, int_point_number, x_int, w_int, connectivity)
            return jacrev(variational_schemes.functional_at_int_point, argnums=3)(
                x_i, w_i, int_point_number, local_dofs, settings, static_settings, set
            )
        residual_at_int_point_vj = vmap(func_at_int_pt, (0,))
        residual_contributions = residual_at_int_point_vj(int_point_numbers)

    elif variational_scheme == "strong form galerkin":
        # Direct implementation of residual, e.g. for Galerkin method
        def func_at_int_pt(int_point_number):
            x_i, w_i, local_dofs = _extract_local_dofs_and_coor_2(dofs, int_point_number, x_int, w_int, connectivity)
            return variational_schemes.direct_residual_at_int_point(
                x_i, w_i, int_point_number, local_dofs, settings, static_settings, set
            )
        residual_at_int_point_vj = vmap(func_at_int_pt, (0,))
        residual_contributions = residual_at_int_point_vj(int_point_numbers)

    elif variational_scheme == "weak form galerkin":
        # Pass local_dofs twice (assuming Bubnov Galerkin...)
        def func_at_int_pt(int_point_number):
            x_i, w_i, local_dofs = _extract_local_dofs_and_coor_2(dofs, int_point_number, x_int, w_int, connectivity)
            return variational_schemes.residual_from_deriv_at_int_point(
                x_i, w_i, int_point_number, local_dofs, local_dofs, settings, static_settings, set
            )
        residual_at_int_point_vj = vmap(func_at_int_pt, (0,))
        residual_contributions = residual_at_int_point_vj(int_point_numbers)

    else:
        raise KeyError("Variational scheme not or wrongly specified!")
    return _get_residual(residual_contributions, connectivity, dofs)

@jit_with_docstring(static_argnames=["static_settings", "set"], possibly_static_argnames=['dofs'])
def sparse_assemble_tangent_diagonal(dofs, settings, static_settings, set):
    """
    Sparse assembly of the diagonal of the tangent matrix for specified set.

    Args:
      dofs (jnp.ndarray or dict): Degrees of freedom.
      settings (dict): Settings dictionary.
      static_settings (flax.core.FrozenDict): Static settings as frozen dictionary.
      set (int): The domain number.

    Returns:
      jnp.ndarray: The diagonal of the assembled tangent matrix.
    """
    connectivity, variational_scheme, x_int, w_int, int_point_numbers = _get_element_quantities_2(dofs, settings, static_settings, set)

    # Compute tangent contributions
    if (
        variational_scheme == "least square pde loss"
        or variational_scheme == "least square function approximation"
    ):
        def func_at_int_pt(int_point_number):
            x_i, w_i, local_dofs = _extract_local_dofs_and_coor_2(dofs, int_point_number, x_int, w_int, connectivity)
            return jacfwd(jacrev(variational_schemes.functional_at_int_point, argnums=3), argnums=3)(
                x_i, w_i, int_point_number, local_dofs, settings, static_settings, set
            )
        at_int_point_vj = vmap(func_at_int_pt, (0,))
        tangent_contributions = at_int_point_vj(int_point_numbers)

    elif variational_scheme == "strong form galerkin":
        def func_at_int_pt(int_point_number):
            x_i, w_i, local_dofs = _extract_local_dofs_and_coor_2(dofs, int_point_number, x_int, w_int, connectivity)
            return jacfwd(variational_schemes.direct_residual_at_int_point, argnums=3)(
                x_i, w_i, int_point_number, local_dofs, settings, static_settings, set
            )
        at_int_point_vj = vmap(func_at_int_pt, (0,))
        tangent_contributions = at_int_point_vj(int_point_numbers)

    elif variational_scheme == "weak form galerkin":
        def func_at_int_pt(int_point_number):
            x_i, w_i, local_dofs = _extract_local_dofs_and_coor_2(dofs, int_point_number, x_int, w_int, connectivity)
            return jacfwd(variational_schemes.residual_from_deriv_at_int_point, argnums=3)(
                x_i, w_i, int_point_number, local_dofs, local_dofs, settings, static_settings, set
            )
        at_int_point_vj = vmap(func_at_int_pt, (0,))
        tangent_contributions = at_int_point_vj(int_point_numbers)

    else:
        raise KeyError("Variational scheme mode not or wrongly specified!")

    return _get_tangent_diagonal(tangent_contributions, connectivity, dofs)

@jit_with_docstring(static_argnames=["static_settings", "set"], possibly_static_argnames=['dofs'])
def sparse_assemble_tangent(dofs, settings, static_settings, set):
    """
    Sparse assembly of the full tangent matrix of specified domain.

    Args:
      dofs (jnp.ndarray or dict): Degrees of freedom.
      settings (dict): Settings dictionary.
      static_settings (flax.core.FrozenDict): Static settings as frozen dictionary.
      set (int): The domain number.

    Returns:
      jax.experimental.sparse.BCOO: The assembled tangent matrix.
    """
    connectivity, variational_scheme, x_int, w_int, int_point_numbers = _get_element_quantities_2(dofs, settings, static_settings, set)

    # Compute tangent contributions
    if (
        variational_scheme == "least square pde loss"
        or variational_scheme == "least square function approximation"
    ):
        def func_at_int_pt(int_point_number):
            x_i, w_i, local_dofs = _extract_local_dofs_and_coor_2(dofs, int_point_number, x_int, w_int, connectivity)
            return jacfwd(jacrev(variational_schemes.functional_at_int_point, argnums=3), argnums=3)(
                x_i, w_i, int_point_number, local_dofs, settings, static_settings, set
            )
        at_int_point_vj = vmap(func_at_int_pt, (0,))
        tangent_contributions = at_int_point_vj(int_point_numbers)

    elif variational_scheme == "strong form galerkin":
        def func_at_int_pt(int_point_number):
            x_i, w_i, local_dofs = _extract_local_dofs_and_coor_2(dofs, int_point_number, x_int, w_int, connectivity)
            return jacfwd(variational_schemes.direct_residual_at_int_point, argnums=3)(
                x_i, w_i, int_point_number, local_dofs, settings, static_settings, set
            )
        at_int_point_vj = vmap(func_at_int_pt, (0,))
        tangent_contributions = at_int_point_vj(int_point_numbers)

    elif variational_scheme == "weak form galerkin":
        def func_at_int_pt(int_point_number):
            x_i, w_i, local_dofs = _extract_local_dofs_and_coor_2(dofs, int_point_number, x_int, w_int, connectivity)
            return jacfwd(variational_schemes.residual_from_deriv_at_int_point, argnums=3)(
                x_i, w_i, int_point_number, local_dofs, local_dofs, settings, static_settings, set
            )
        at_int_point_vj = vmap(func_at_int_pt, (0,))
        tangent_contributions = at_int_point_vj(int_point_numbers)

    else:
        raise KeyError("Variational scheme mode not or wrongly specified!")

    # Assembling (without summing duplicates)
    data = dict_flatten(tangent_contributions)
    indices = _get_indices(connectivity, dofs)
    num_dofs = (
        dofs.size if not isinstance(dofs, dict) else sum(v.size for v in dofs.values())
    )
    tangent_matrix = sparse.BCOO((data, indices), shape=(num_dofs, num_dofs))
    return tangent_matrix


### Assembling for user potentials
@jit_with_docstring(static_argnames=["static_settings", "set"], possibly_static_argnames=['dofs'])
def user_potential_integrate_functional(dofs, settings, static_settings, set):
    """
    Assembly of potential for custom user definition of specified domain.

    Args:
      dofs (jnp.ndarray or dict or callable): Degrees of freedom. Can be a function of time for transient problems.
      settings (dict): Settings dictionary.
      static_settings (flax.core.FrozenDict): Static settings as frozen dictionary.
      set (int): The domain number.

    Returns:
      float: value of functional integrated over set of elements
    """
    model_fun, x_nodes, elem_numbers, connectivity = _get_element_quantities(dofs, settings, static_settings, set)

    # Modify the model_fun such that it extracts the DOFs from the global dofs and vmap only over connectivity
    def element_residual(elem_number, node_list):
        local_dofs, local_node_coor = _extract_local_dofs_and_coor(dofs, node_list, x_nodes)
        return  model_fun(local_dofs, local_node_coor, elem_number, settings, static_settings, set)

    functional_contributions = vmap(element_residual, (0, 0), (0))(elem_numbers, connectivity)

    return functional_contributions.sum()

@jit_with_docstring(static_argnames=["static_settings", "set"], possibly_static_argnames=['dofs'])
def user_potential_assemble_residual(dofs, settings, static_settings, set):
    """
    Assembly of residual for custom user potential of specified domain.

    Args:
      dofs (jnp.ndarray or dict): Degrees of freedom.
      settings (dict): Settings dictionary.
      static_settings (flax.core.FrozenDict): Static settings as frozen dictionary.
      set (int): The domain number.

    Returns:
      jnp.ndarray: The assembled residual.
    """
    model_fun, x_nodes, elem_numbers, connectivity = _get_element_quantities(dofs, settings, static_settings, set)

    # Modify the model_fun such that it extracts the DOFs from the global dofs and vmap only over connectivity
    def element_residual(elem_number, node_list):
        local_dofs, local_node_coor = _extract_local_dofs_and_coor(dofs, node_list, x_nodes)
        return  jacrev(model_fun)(local_dofs, local_node_coor, elem_number, settings, static_settings, set)

    # residual_contributions = vmap(element_residual, (0, 0))(elem_numbers, connectivity)
    residual_contributions = _batched_map(element_residual, elem_numbers, connectivity)

    return _get_residual(residual_contributions, connectivity, dofs)

@jit_with_docstring(static_argnames=["static_settings", "set"], possibly_static_argnames=['dofs'])
def user_potential_assemble_tangent_diagonal(dofs, settings, static_settings, set):
    """
    Assembly of the diagonal of the tangent matrix for custom user potential of specified domain.

    Args:
      dofs (jnp.ndarray or dict): Degrees of freedom.
      settings (dict): Settings dictionary.
      static_settings (flax.core.FrozenDict): Static settings as frozen dictionary.
      set (int): The domain number.

    Returns:
      jnp.ndarray: The diagonal of the assembled tangent matrix.
    """
    model_fun, x_nodes, elem_numbers, connectivity = _get_element_quantities(dofs, settings, static_settings, set)

    # Modify the model_fun such that it extracts the dofs from the global dofs and vmap only over connectivity
    def element_tangent(elem_number, node_list):
        local_dofs, local_node_coor = _extract_local_dofs_and_coor(dofs, node_list, x_nodes)
        return  jacfwd(jacrev(model_fun))(local_dofs, local_node_coor, elem_number, settings, static_settings, set)

    tangent_contributions = vmap(element_tangent, (0, 0), (0))(elem_numbers, connectivity)

    return _get_tangent_diagonal(tangent_contributions, connectivity, dofs)

@jit_with_docstring(static_argnames=["static_settings", "set"], possibly_static_argnames=['dofs'])
def user_potential_assemble_tangent(dofs, settings, static_settings, set):
    """
    Assembly of the full (sparse) tangent matrix for custom user potential of specified domain.

    Args:
      dofs (jnp.ndarray or dict): Degrees of freedom.
      settings (dict): Settings dictionary.
      static_settings (flax.core.FrozenDict): Static settings as frozen dictionary.
      set (int): The domain number.

    Returns:
      jax.experimental.sparse.BCOO: The assembled tangent matrix.
    """
    model_fun, x_nodes, elem_numbers, connectivity = _get_element_quantities(dofs, settings, static_settings, set)

    # Modify the model_fun such that it extracts the dofs from the global dofs and vmap only over connectivity
    def element_tangent(elem_number, node_list):
        local_dofs, local_node_coor = _extract_local_dofs_and_coor(dofs, node_list, x_nodes)
        return  jacfwd(jacrev(model_fun))(local_dofs, local_node_coor, elem_number, settings, static_settings, set)

    # tangent_contributions = vmap(element_tangent, (0, 0), (0))(elem_numbers, connectivity)

    body_fun = lambda i: element_tangent(elem_numbers[i], jax.tree.map(lambda x: x[i], connectivity))
    num_dofs_per_elem = jax.eval_shape(lambda i: dict_flatten(body_fun(i)), 0).shape[0]
    tangent_contributions = jax.lax.map(body_fun, jnp.arange(elem_numbers.shape[0]), batch_size=int(64000/num_dofs_per_elem))

    data = dict_flatten(tangent_contributions)
    indices = _get_indices(connectivity, dofs)
    num_dofs = (
        dofs.size if not isinstance(dofs, dict) else sum(v.size for v in dofs.values())
    )

    tangent_matrix = sparse.BCOO((data, indices), shape=(num_dofs, num_dofs))
    return tangent_matrix

@jit_with_docstring(static_argnames=["static_settings", "set"], possibly_static_argnames=['dofs'])
def _user_potential_assemble_r_and_t(dofs, settings, static_settings, set):
    model_fun, x_nodes, elem_numbers, connectivity = _get_element_quantities(dofs, settings, static_settings, set)

    # Modify the model_fun such that it extracts the dofs from the global dofs and vmap only over connectivity
    def element_r_and_t(elem_number, node_list):
        local_dofs, local_node_coor = _extract_local_dofs_and_coor(dofs, node_list, x_nodes)
        residual_fun = lambda x: jacrev(model_fun)(x, local_node_coor, elem_number, settings, static_settings, set)

        elem_res = residual_fun(local_dofs)
        elem_tan = jacfwd(residual_fun)(local_dofs)

        # def residual_and_tangent_linearize(residual_fun, local_dofs):
        #     primals, lin_fun = linearize(residual_fun, local_dofs)            
        #     flat_local_dofs = dict_flatten(local_dofs)
        #     n = flat_local_dofs.shape[0]            
        #     identity = jnp.eye(n)            
        #     jacobian = vmap(lambda v: lin_fun(reshape_as(v, local_dofs)))(identity)            
        #     return primals, jacobian
        # elem_res, elem_tan = residual_and_tangent_linearize(residual_fun, local_dofs)

        # def residual_and_tangent(residual_fun, local_dofs):
        #     flat_local_dofs = dict_flatten(local_dofs)
        #     n = flat_local_dofs.shape[0]
        #     identity = jnp.eye(n)
        #     def jvp_with_flat_tangent(v):
        #         tangent_pytree = reshape_as(v, local_dofs)
        #         return jvp(residual_fun, (local_dofs,), (tangent_pytree,))
        #     primals, elem_tan = vmap(jvp_with_flat_tangent)(identity)
        #     elem_res = treemap(lambda x: x[0], primals)
        #     return elem_res, elem_tan        
        # elem_res, elem_tan = residual_and_tangent(residual_fun, local_dofs)

        # def residual_and_tangent_vjp(residual_fun, local_dofs):
        #     primals, vjp_fun = vjp(residual_fun, local_dofs)          
        #     flat_res = dict_flatten(primals)
        #     m = flat_res.shape[0]
        #     flat_local = dict_flatten(local_dofs)
        #     n = flat_local.shape[0]
        #     identity = jnp.eye(m)
        #     jacobian_rows = vmap(
        #         lambda v: dict_flatten(vjp_fun(reshape_as(v, primals))[0])
        #     )(identity)
        #     jacobian = jacobian_rows.T
        #     return primals, jacobian
        # elem_res, elem_tan = residual_and_tangent_vjp(residual_fun, local_dofs)

        return elem_res, elem_tan

    all_contributions = vmap(element_r_and_t, (0, 0), (0, 0))
    residual_contributions, tangent_contributions = all_contributions(elem_numbers, connectivity)

    residual = _get_residual(residual_contributions, connectivity, dofs)

    data = dict_flatten(tangent_contributions)
    indices = _get_indices(connectivity, dofs)
    num_dofs = (
        dofs.size if not isinstance(dofs, dict) else sum(v.size for v in dofs.values())
    )

    tangent_matrix = sparse.BCOO((data, indices), shape=(num_dofs, num_dofs))
    return residual, tangent_matrix


### Assembling for user residuals
@jit_with_docstring(static_argnames=["static_settings", "set"], possibly_static_argnames=['dofs'])
def user_residual_assemble_residual(dofs, settings, static_settings, set):
    """
    Assembly of residual for custom user residual of specified domain.

    Args:
      dofs (jnp.ndarray or dict or callable): Degrees of freedom. Can be a function of time for transient problems.
      settings (dict): Settings dictionary.
      static_settings (flax.core.FrozenDict): Static settings as frozen dictionary.
      set (int): The domain number.

    Returns:
      jnp.ndarray: The assembled residual.
    """
    model_fun, x_nodes, elem_numbers, connectivity = _get_element_quantities(dofs, settings, static_settings, set)

    # Modify the model_fun such that it extracts the DOFs from the global dofs and vmap only over connectivity
    def element_residual(elem_number, node_list):
        local_dofs, local_node_coor = _extract_local_dofs_and_coor(dofs, node_list, x_nodes)
        return  model_fun(local_dofs, local_node_coor, elem_number, settings, static_settings, set)
    
    # residual_contributions = vmap(element_residual, (0, 0), (0))(elem_numbers, connectivity)

    body_fun = lambda i: element_residual(elem_numbers[i], jax.tree.map(lambda x: x[i], connectivity))
    num_dofs_per_elem = jax.eval_shape(lambda i: dict_flatten(body_fun(i)), 0).shape[0]
    residual_contributions = jax.lax.map(body_fun, jnp.arange(elem_numbers.shape[0]), batch_size=int(64000/num_dofs_per_elem))

    return _get_residual(residual_contributions, connectivity, dofs(0.) if callable(dofs) else dofs)

@jit_with_docstring(static_argnames=["static_settings", "set"], possibly_static_argnames=['dofs'])
def user_residual_assemble_tangent_diagonal(dofs, settings, static_settings, set):
    """
    Assembly of the diagonal of the tangent matrix for custom user residual of specified domain.

    Args:
      dofs (jnp.ndarray or dict): Degrees of freedom.
      settings (dict): Settings dictionary.
      static_settings (flax.core.FrozenDict): Static settings as frozen dictionary.
      set (int): The domain number.

    Returns:
      jnp.ndarray: The diagonal of the assembled tangent matrix.
    """
    model_fun, x_nodes, elem_numbers, connectivity = _get_element_quantities(dofs, settings, static_settings, set)

    # Modify the model_fun such that it extracts the DOFs from the global dofs and vmap only over connectivity
    def element_residual(elem_number, node_list):
        local_dofs, local_node_coor = _extract_local_dofs_and_coor(dofs, node_list, x_nodes)
        return  jacfwd(model_fun)(local_dofs, local_node_coor, elem_number, settings, static_settings, set)

    tangent_contributions = vmap(element_residual, (0, 0), (0))(elem_numbers, connectivity)

    return _get_tangent_diagonal(tangent_contributions, connectivity, dofs)

@jit_with_docstring(static_argnames=["static_settings", "set"], possibly_static_argnames=['dofs'])
def user_residual_assemble_tangent(dofs, settings, static_settings, set):
    """
    Assembly of the full (sparse) tangent matrix for custom user residual of specified domain.

    Args:
      dofs (jnp.ndarray or dict): Degrees of freedom.
      settings (dict): Settings dictionary.
      static_settings (flax.core.FrozenDict): Static settings as frozen dictionary.
      set (int): The domain number.

    Returns:
      jax.experimental.sparse.BCOO: The assembled tangent matrix.
    """
    model_fun, x_nodes, elem_numbers, connectivity = _get_element_quantities(dofs, settings, static_settings, set)

    # Modify the model_fun such that it extracts the DOFs from the global dofs and vmap only over connectivity
    def element_residual(elem_number, node_list):
        local_dofs, local_node_coor = _extract_local_dofs_and_coor(dofs, node_list, x_nodes)
        return  jacfwd(model_fun)(local_dofs, local_node_coor, elem_number, settings, static_settings, set)

    tangent_contributions = vmap(element_residual, (0, 0), (0))(elem_numbers, connectivity)

    data = dict_flatten(tangent_contributions)
    indices = _get_indices(connectivity, dofs)

    if callable(dofs):
        dofs0 = dofs(0.)
        num_dofs = (
        dofs0.size if not isinstance(dofs0, dict) else sum(v.size for v in dofs0.values())
        )
    else:
        num_dofs = (
            dofs.size if not isinstance(dofs, dict) else sum(v.size for v in dofs.values())
        )
    tangent_matrix = sparse.BCOO((data, indices), shape=(num_dofs, num_dofs))
    return tangent_matrix


### Assembling for user elements
@jit_with_docstring(static_argnames=["static_settings", "set"], possibly_static_argnames=['dofs'])
def user_element_assemble_residual(dofs, settings, static_settings, set):
    """
    Assembly of residual for custom user element of specified domain.

    Args:
      dofs (jnp.ndarray or dict): Degrees of freedom.
      settings (dict): Settings dictionary.
      static_settings (flax.core.FrozenDict): Static settings as frozen dictionary.
      set (int): The domain number.

    Returns:
      jnp.ndarray: The assembled residual.
    """
    model_fun, x_nodes, elem_numbers, connectivity = _get_element_quantities(dofs, settings, static_settings, set)

    # Modify the model_fun such that it extracts the DOFs from the global dofs and vmap only over connectivity
    def element_residual(elem_number, node_list):
        local_dofs, local_node_coor = _extract_local_dofs_and_coor(dofs, node_list, x_nodes)
        return  model_fun(local_dofs, local_node_coor, elem_number, settings, static_settings, "residual", set)

    residual_contributions = vmap(element_residual, (0, 0), (0))(elem_numbers, connectivity)

    return _get_residual(residual_contributions, connectivity, dofs)

@jit_with_docstring(static_argnames=["static_settings", "set"], possibly_static_argnames=['dofs'])
def user_element_assemble_tangent_diagonal(dofs, settings, static_settings, set):
    """
    Assembly of the diagonal of the tangent matrix for custom user element of specified domain.

    Args:
      dofs (jnp.ndarray or dict): Degrees of freedom.
      settings (dict): Settings dictionary.
      static_settings (flax.core.FrozenDict): Static settings as frozen dictionary.
      set (int): The domain number.

    Returns:
      jnp.ndarray: The diagonal of the assembled tangent matrix.
    """
    model_fun, x_nodes, elem_numbers, connectivity = _get_element_quantities(dofs, settings, static_settings, set)

    # Modify the model_fun such that it extracts the DOFs from the global dofs and vmap only over connectivity
    def element_residual(elem_number, node_list):
        local_dofs, local_node_coor = _extract_local_dofs_and_coor(dofs, node_list, x_nodes)
        return  model_fun(local_dofs, local_node_coor, elem_number, settings, static_settings, "tangent", set)

    tangent_contributions = vmap(element_residual, (0, 0), (0))(elem_numbers, connectivity)

    return _get_tangent_diagonal(tangent_contributions, connectivity, dofs)

@jit_with_docstring(static_argnames=["static_settings", "set"], possibly_static_argnames=['dofs'])
def user_element_assemble_tangent(dofs, settings, static_settings, set):
    """
    Assembly of the full (sparse) tangent matrix for custom user element of specified domain.

    Args:
      dofs (jnp.ndarray or dict): Degrees of freedom.
      settings (dict): Settings dictionary.
      static_settings (flax.core.FrozenDict): Static settings as frozen dictionary.
      set (int): The domain number.

    Returns:
      jax.experimental.sparse.BCOO: The assembled tangent matrix.
    """
    model_fun, x_nodes, elem_numbers, connectivity = _get_element_quantities(dofs, settings, static_settings, set)

    # Modify the model_fun such that it extracts the DOFs from the global dofs and vmap only over connectivity
    def element_residual(elem_number, node_list):
        local_dofs, local_node_coor = _extract_local_dofs_and_coor(dofs, node_list, x_nodes)
        return  model_fun(local_dofs, local_node_coor, elem_number, settings, static_settings, "tangent", set)

    tangent_contributions = vmap(element_residual, (0, 0))(elem_numbers, connectivity)

    data = dict_flatten(tangent_contributions)
    indices = _get_indices(connectivity, dofs)
    num_dofs = (
        dofs.size if not isinstance(dofs, dict) else sum(v.size for v in dofs.values())
    )
    tangent_matrix = sparse.BCOO((data, indices), shape=(num_dofs, num_dofs))
    return tangent_matrix


### Internal variable update function
@jit_with_docstring(static_argnames=["static_settings", "set"], possibly_static_argnames=['dofs'])
def get_int_var_updates(dofs, settings, static_settings, set):
    """
    Get the internal variables for a specified domain for all elements and integration points.

    Similar Structure as the assemble_residual functions, but uses 'int var updates' instead of 'model' in order to compute the per element and Gauss point internal variables.
    """
    _, x_nodes, elem_numbers, connectivity = _get_element_quantities(dofs, settings, static_settings, set)
    int_var_updates = static_settings['int var updates'][set]
    def local_update_fun(elem_number, node_list):
        local_dofs, local_node_coor = _extract_local_dofs_and_coor(dofs, node_list, x_nodes)
        return  int_var_updates(local_dofs, local_node_coor, elem_number, settings, static_settings, set)
    return jax.vmap(local_update_fun, (0, 0), (0))(elem_numbers, connectivity)



def test_example():

  # Imports
  import jax
  import jax.numpy as jnp
  import flax

  from autopdex import seeder, geometry, solver, utility, models, spaces, mesher
  jax.config.update("jax_enable_x64", True)

  # Generate mesh (or import the node coordinates and connectivity)
  pts = [[0., 0.], [1., 0.], [1., 1.], [0., 1.]]
  coords, elems = mesher.structured_mesh((5, 5), pts, 'quad')
  node_coordinates = {'phi': coords,}
  connectivity = {'phi': elems,}

  # Selection of nodes and degrees of freedom for Dirichlet conditions
  sdf = lambda x: geometry.psdf_polygon(x, pts)
  dirichlet_nodes = geometry.in_sdfs(node_coordinates['phi'], sdf)
  dirichlet_dofs = {'phi': dirichlet_nodes,}
  dirichlet_conditions = utility.dict_zeros_like(dirichlet_dofs, dtype=jnp.float64)

  # Define the variational problem
  def integrand_fun(x_int, ansatz_fun, settings, static_settings, elem_number, set):
      # Definition of custom functional    
      x = ansatz_fun['physical coor'](x_int)
      phi_fun = ansatz_fun['phi']
      phi = phi_fun(x_int)
      dphi_dx = jax.jacrev(phi_fun)(x_int)

      x_1 = x
      x_2 = x - jnp.array([1., 0.5])
      source_term = 20 * (jnp.sin(10 * x_1 @ x_1) - jnp.cos(10 * x_2 @ x_2))
      return (1/2) * dphi_dx @ dphi_dx - source_term * phi

  # Set up the finite element, here Q1 elements for the field 'phi'
  user_potential = models.mixed_reference_domain_potential(
      integrand_fun,
      {'phi': spaces.fem_iso_line_quad_brick,},
      *seeder.gauss_legendre_nd(dimension = 2, order = 2),
      'phi')

  # Prepare the settings for autopdex
  static_settings = flax.core.FrozenDict({
    'assembling mode': ('user potential',),
    'solution structure': ('nodal imposition',),
    'model': (user_potential, ),
    'solver type': 'newton',
    'solver backend': 'scipy',
    'solver': 'lapack',
    'verbose': -1,
  })
  settings = {
    'connectivity': (connectivity,),
    'dirichlet dofs': dirichlet_dofs,
    'node coordinates': node_coordinates,
    'dirichlet conditions': dirichlet_conditions,
  }

  # Compile, assemble and solve linear system
  initial_guess = utility.dict_zeros_like(dirichlet_dofs, dtype=jnp.float64)
  test = solver.solver(initial_guess, settings, static_settings)[0]['phi'].sum()
  assert jnp.isclose(test, 1.9066412530282952)

# test_example()


# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.


"""
These functions are a modification and extension of the functions in jaxopt._src.implicit_diff for
external solve functions with a jax.experimental.sparse.BCOO matrix as an argument instead of a matvec function.
The root_vjp and root_jvp functions were modified in a way that external solvers 
can be used via a pure_callback and constraints can be taken into account.
With the wrapper custom_root, root solvers can be made differentiable both in forward or reverse mode of arbitrary order.
Mixing the differentiation mode is currently not possible.
"""

import inspect
from typing import Any
from typing import Callable
from typing import Optional
from typing import Tuple

import jax
import jax.numpy as jnp

from autopdex import utility


def tree_scalar_mul(scalar, tree_x):
    """Compute scalar * tree_x."""
    return jax.tree_util.tree_map(lambda x: scalar * x, tree_x)

def _extract_kwargs(kwarg_keys, flat_args):
    n = len(flat_args) - len(kwarg_keys)
    args, kwarg_vals = flat_args[:n], flat_args[n:]
    kwargs = dict(zip(kwarg_keys, kwarg_vals))
    return args, kwargs

def _signature_bind(signature, *args, **kwargs):
    ba = signature.bind(*args, **kwargs)
    ba.apply_defaults()
    return ba.args, ba.kwargs

def _signature_bind_and_match(signature, *args, **kwargs):
    # We want to bind *args and **kwargs based on the provided
    # signature, but also to associate the resulting positional
    # arguments back. To achieve this, we lift arguments to a triple:
    #
    #   (was_kwarg, ref, value)
    #
    # where ref is an index position (int) if the original argument was
    # from *args and a dictionary key if the original argument was from
    # **kwargs. After binding to the inspected signature, we use the
    # tags to associate the resolved positional arguments back to their
    # arg and kwarg source.

    args = [(False, i, v) for i, v in enumerate(args)]
    kwargs = {k: (True, k, v) for (k, v) in kwargs.items()}
    ba = signature.bind(*args, **kwargs)

    mapping = [(was_kwarg, ref) for was_kwarg, ref, _ in ba.args]

    def map_back(out_args):
        src_args = [None] * len(args)
        src_kwargs = {}
        for (was_kwarg, ref), out_arg in zip(mapping, out_args):
            if was_kwarg:
                src_kwargs[ref] = out_arg
            else:
                src_args[ref] = out_arg
        return src_args, src_kwargs

    out_args = tuple(v for _, _, v in ba.args)
    out_kwargs = {k: v for k, (_, _, v) in ba.kwargs.items()}
    return out_args, out_kwargs, map_back

def _jvp_args(residual_fun, sol, args, tangents):
    """JVP in the second argument of residual_fun."""
    # We close over the solution.
    fun = lambda *y: residual_fun(sol, *y)
    return jax.jvp(fun, args, tangents)[1]

def _root_vjp(
    residual_fun: Callable,
    mat_fun: Callable,
    sol: Any,
    args: Tuple,
    cotangent: Any,
    solve_fun: Callable,
    free_dofs: Any,
) -> Any:
    """Vector-Jacobian product of a root.

    The invariant is ``residual_fun(sol, *args) == 0``.

    Args:
      residual_fun: the optimality function to use.
      sol: solution / root (pytree).
      mat_fun: a function that has to compute the sparse tangent matrix with sol and args as arguments.
      args: tuple containing the arguments with respect to which we wish to
        differentiate ``sol`` against.
      cotangent: vector to left-multiply the Jacobian with
        (pytree, same structure as ``sol``).
      solve_fun: a linear solver of the form ``x = solve_fun(mat, b)``,
        where ``mat`` is as jax.experimental.sparse.BCOO matrix.
    Returns:
      tuple of the same length as ``len(args)`` containing the vjps w.r.t.
      each argument. Each ``vjps[i]`` has the same pytree structure as
      ``args[i]``.
    """

    free_dofs_flat = None
    if free_dofs is not None:
        assert "dirichlet conditions" in args[0], "'dirichlet conditions' \
        have to be defined in a dict as the second argument of the root solver function."
        dirichlet_dofs_flat = utility.dict_flatten(args[0]["dirichlet dofs"])
        free_dofs_flat = ~dirichlet_dofs_flat

    @jax.custom_vjp
    def linear_solver_fun_vjp(A, b, free_dofs_flat):
        return solve_fun(A, b, free_dofs_flat)  # Here can be an external callback

    def linear_solver_fun_vjp_fwd(A, b, free_dofs_flat):
        result = linear_solver_fun_vjp(A, b, free_dofs_flat)
        return result, (A, b, free_dofs_flat, result)

    def linear_solver_fun_vjp_bwd(res, g):
        A, b, free_dofs_flat, result = res

        # Ensure g matches the size of the expected gradient
        if free_dofs is not None:

            # Sparse outer product
            result_dot = linear_solver_fun_vjp(A.T, g, free_dofs_flat)

            size = A.shape[0]
            empty = jnp.zeros((size,), dtype=float)
            result_tmp = utility.mask_op(empty, free_dofs_flat, result, "set")
            result_dot_tmp = utility.mask_op(empty, free_dofs_flat, result_dot, "set")

            # # Use the sparsity structure of A to build Fx_dot
            indices = jnp.asarray(A.indices, dtype=int)
            data = -result_dot_tmp[indices[:, 0]] * result_tmp[indices[:, 1]]

            # Construct Fx_dot with the same sparsity pattern
            Fx_dot = jax.experimental.sparse.BCOO((data, indices), shape=A.shape)
            Fy_dot = result_dot
        else:
            # Sparse outer product
            result_dot = linear_solver_fun_vjp(A.T, g, None)

            # Use the sparsity structure of A to build Fx_dot
            indices = jnp.asarray(A.indices, dtype=int)
            data = -result_dot[indices[:, 0]] * result[indices[:, 1]]

            # Construct Fx_dot with the same sparsity pattern
            Fx_dot = jax.experimental.sparse.BCOO((data, indices), shape=A.shape)
            Fy_dot = result_dot

        # Correctly return the shapes that match the input arguments
        return (Fx_dot, Fy_dot, None)

    linear_solver_fun_vjp.defvjp(linear_solver_fun_vjp_fwd, linear_solver_fun_vjp_bwd)
    diffable_solve_fun = linear_solver_fun_vjp

    mat = mat_fun(sol, *args)

    # The solution of A^T u = v, where
    # A = jacobian(residual_fun, argnums=0)
    # v = -cotangent.
    v = tree_scalar_mul(-1, cotangent)

    if free_dofs is not None:
        dirichlet_dofs = utility.reshape_as(dirichlet_dofs_flat, sol)

        v_flat = utility.dict_flatten(v)
        u_f = diffable_solve_fun(mat.T, v_flat, free_dofs_flat)
        u_flat = utility.mask_op(
            utility.dict_flatten(utility.dict_zeros_like(v)), free_dofs_flat, u_f, "set"
        )

        def fun_args(*args):
            def residual_fun_tmp(sol, *args):
                dirichlet_values = args[0]["dirichlet conditions"]
                sol_with_bc = utility.mask_op(
                    sol, dirichlet_dofs, dirichlet_values, "set"
                )
                return residual_fun(sol_with_bc, *args)

            return residual_fun_tmp(sol, *args)

        u = utility.reshape_as(u_flat, v)
    else:

        def fun_args(*args):
            return residual_fun(sol, *args)

        v_flat = utility.dict_flatten(v)
        u = utility.reshape_as(diffable_solve_fun(mat.T, v_flat, None), v)

    _, vjp_fun_args = jax.vjp(fun_args, *args)
    args_vjp = vjp_fun_args(u)

    if free_dofs is not None:
        updated_args0 = args_vjp[0]
        tmpl = updated_args0["dirichlet conditions"]
        updated_args0["dirichlet conditions"] = utility.mask_op(
            updated_args0["dirichlet conditions"], utility.reshape_as(dirichlet_dofs, tmpl), utility.reshape_as(cotangent, tmpl), "add"
        )
        args_vjp = (updated_args0,) + args_vjp[1:]

    return args_vjp

def _root_jvp(
    residual_fun: Callable,
    mat_fun: Callable,
    sol: Any,
    args: Tuple,
    tangents: Tuple,
    solve_fun: Callable,
    free_dofs: Any,
    # dirichlet_dofs: Any,
) -> Any:
    """
    Jacobian-vector product of a root.

    The invariant is ``residual_fun(sol, *args) == 0``.

    Args:
      residual_fun: the optimality function to use.
      mat_fun: a function that has to compute the sparse tangent matrix with sol and args as arguments.
      sol: solution / root (pytree).
      args: tuple containing the arguments with respect to which to differentiate.
      tangents: a tuple of the same size as ``len(args)``. Each ``tangents[i]``
        has the same pytree structure as ``args[i]``.
      solve_fun: a linear solver of the form ``x = solve_fun(mat, b)``,
        where ``mat`` is as jax.experimental.sparse.BCOO matrix.
    Returns:
      a pytree with the same structure as ``sol``.
    """
    free_dofs_flat = None
    if free_dofs is not None:
        assert "dirichlet conditions" in args[0], "'dirichlet conditions' \
        have to be defined in a dict as the second argument of the root solver function."
        dirichlet_dofs_flat = utility.dict_flatten(args[0]["dirichlet dofs"])
        free_dofs_flat = ~dirichlet_dofs_flat

    # Compute tangent matrix
    A = mat_fun(sol, *args)
    mat_shape = A.shape

    # Forward differentiable sparse linear solver
    # TODO: register as primitive in order to allow mixed jacfwd/jacrev
    @jax.custom_jvp
    def linear_solver_fun_jvp(data, indices, b, free_dofs_flat):
        A = jax.experimental.sparse.BCOO((data, indices), shape=mat_shape)
        return solve_fun(A, b, free_dofs_flat)

    @linear_solver_fun_jvp.defjvp
    def linear_solver_fun_jvp_rule(primals, tangents):
        data, indices, b, free_dofs_flat = primals
        data_dot, _, b_dot, _ = tangents

        # Compute the primal result using the linear solver function
        primal_result = linear_solver_fun_jvp(data, indices, b, free_dofs_flat)

        # ToDo: is it somehow possible without A_dot via jvps?
        A_dot = jax.experimental.sparse.BCOO((data_dot, indices), shape=mat_shape)

        # Handle the tangent calculation
        if free_dofs is not None:
            primal_result_tmp = utility.mask_op(
                jnp.zeros((mat_shape[0],), dtype=float),
                free_dofs_flat,
                primal_result,
            )
            rhs = b_dot - (A_dot @ primal_result_tmp)
            result_dot = linear_solver_fun_jvp(data, indices, rhs, free_dofs_flat)
        else:
            result_dot = linear_solver_fun_jvp(
                data, indices, b_dot - A_dot @ primal_result, None
            )

        return primal_result, result_dot

    # Assign the jvp-enabled solver function
    solve_func = linear_solver_fun_jvp

    if free_dofs is not None:
        dirichlet_dofs = utility.reshape_as(dirichlet_dofs_flat, sol)

        # Explicit imposition of DOFs in order to be able to take derivatives w.r.t. nodally imposed DOFs
        def residual_fun_tmp(sol, *args):
            sol_with_bc = utility.mask_op(
                sol, dirichlet_dofs, args[0]["dirichlet conditions"], "set"
            )
            return residual_fun(sol_with_bc, *args)

        Bv = _jvp_args(residual_fun_tmp, sol, args, tangents)
        Bv_free = utility.dict_flatten(Bv)
        Jv_free = solve_func(A.data, A.indices, -Bv_free, free_dofs_flat)

        empty_flat = utility.dict_flatten(utility.dict_zeros_like(sol))
        Jv = utility.reshape_as(
            utility.mask_op(empty_flat, free_dofs_flat, Jv_free, "set"), sol
        )

        Jv = utility.mask_op(
            Jv, dirichlet_dofs, tangents[0]["dirichlet conditions"], "set"
        )

    else:
        Bv = _jvp_args(residual_fun, sol, args, tangents)
        Jv = utility.reshape_as(
            solve_func(A.data, A.indices, -utility.dict_flatten(Bv), None), Bv
        )

    return Jv

def _custom_root(
    solver_fun,
    residual_fun,
    mat_fun,
    free_dofs,
    solve,
    has_aux,
    mode="reverse",
    reference_signature=None,
):
    # When caling through `jax.custom_vjp`, jax attempts to resolve all
    # arguments passed by keyword to positions (this is in order to
    # match against a `nondiff_argnums` parameter that we do not use
    # here). It does so by resolving them according to the custom_jvp'ed
    # function's signature. It disallows functions defined with a
    # catch-all `**kwargs` expression, since their signature cannot
    # always resolve all keyword arguments to positions.
    #
    # We can loosen the constraint on the signature of `solver_fun` so
    # long as we resolve keywords to positions ourselves. We can do so
    # just in time, by flattening the `kwargs` dict (respecting its
    # iteration order) and supplying `custom_vjp` with a
    # positional-argument-only function. We then explicitly coordinate
    # flattening and unflattening around the `custom_vjp` boundary.
    #
    # Once we make it past the `custom_vjp` boundary, we do some more
    # work to align arguments with the reference signature (which is, by
    # default, the signature of `residual_fun`).

    solver_fun_signature = inspect.signature(solver_fun)

    if reference_signature is None:
        reference_signature = inspect.signature(residual_fun)

    elif not isinstance(reference_signature, inspect.Signature):
        # If is a CompositeLinearFunction, accesses subfun.
        # Otherwise, assumes a Callable.
        fun = getattr(reference_signature, "subfun", reference_signature)
        reference_signature = inspect.signature(fun)

    def make_custom_solver_fun(solver_fun, kwarg_keys):

        def solver_fun_flat_tmp(*flat_args):
            args, kwargs = _extract_kwargs(kwarg_keys, flat_args)
            return solver_fun(*args, **kwargs)

        if mode == "reverse" or mode == "backward":
            solver_fun_flat = jax.custom_vjp(solver_fun_flat_tmp)
        elif mode == "forward":
            solver_fun_flat = jax.custom_jvp(solver_fun_flat_tmp)
        else:
            raise ValueError("Mode must be either 'forward' or 'reverse'.")

        # Forward-mode differentiation (JVP)
        def solver_fun_jvp(primals, tangents):
            args, kwargs = _extract_kwargs(kwarg_keys, primals)
            tangent_args, tangent_kwargs = _extract_kwargs(kwarg_keys, tangents)

            # Compute the primal solution using the root solver function
            primal_sol = solver_fun_flat(*args, **kwargs)

            # Handle has_aux case
            if has_aux:
                sol = primal_sol[0]
                aux_data = primal_sol[1:]

                # TODO: allow integer and boolean auxilary data
            else:
                sol = primal_sol

            # Compute JVP using root_jvp
            jvp_sol = _root_jvp(
                residual_fun=residual_fun,
                mat_fun=mat_fun,
                sol=sol,
                args=args[1:],  # Exclude the initial params from args
                tangents=tangent_args[1:],  # Exclude the initial params from tangents
                solve_fun=solve,
                free_dofs=free_dofs,
            )

            if has_aux:
                # Return primal and tangent for both solution and auxiliary data
                aux_tangent = jax.tree_util.tree_map(jnp.zeros_like, aux_data)
                return (sol,) + aux_data, (jvp_sol,) + aux_tangent
            else:
                return primal_sol, jvp_sol

        # Reverse-mode differentiation (VJP)
        def solver_fun_fwd(*flat_args):
            res = solver_fun_flat(*flat_args)
            return res, (res, flat_args)

        def solver_fun_bwd(tup, cotangent):
            res, flat_args = tup
            args, kwargs = _extract_kwargs(kwarg_keys, flat_args)

            # solver_fun can return auxiliary data if has_aux = True.
            if has_aux:
                cotangent = cotangent[0]
                sol = res[0]
            else:
                sol = res

            ba_args, ba_kwargs, map_back = _signature_bind_and_match(
                reference_signature, *args, **kwargs
            )
            if ba_kwargs:
                raise TypeError(
                    "keyword arguments to solver_fun could not be resolved to "
                    "positional arguments based on the signature "
                    f"{reference_signature}. This can happen under custom_root if "
                    "residual_fun takes catch-all **kwargs, or under "
                    "custom_fixed_point if fixed_point_fun takes catch-all **kwargs, "
                    "both of which are currently unsupported."
                )

            # Compute VJPs w.r.t. args.
            vjps = _root_vjp(
                residual_fun=residual_fun,
                mat_fun=mat_fun,
                sol=sol,
                args=ba_args[1:],
                cotangent=cotangent,
                solve_fun=solve,
                free_dofs=free_dofs,
            )
            # Prepend None as the vjp for init_params.
            vjps = (None,) + vjps

            arg_vjps, kws_vjps = map_back(vjps)
            ordered_vjps = tuple(arg_vjps) + tuple(kws_vjps[k] for k in kwargs.keys())
            return ordered_vjps

        if mode == "reverse" or mode == "backward":
            solver_fun_flat.defvjp(solver_fun_fwd, solver_fun_bwd)
        else:
            solver_fun_flat.defjvp(solver_fun_jvp)
        return solver_fun_flat

    def wrapped_solver_fun(*args, **kwargs):
        args, kwargs = _signature_bind(solver_fun_signature, *args, **kwargs)
        keys, vals = list(kwargs.keys()), list(kwargs.values())
        return make_custom_solver_fun(solver_fun, keys)(*args, *vals)

    return wrapped_solver_fun

def custom_root(
    residual_fun: Callable,
    mat_fun: Callable,
    solve: Callable,
    free_dofs = None,
    has_aux: bool = False,
    mode="reverse",
    reference_signature: Optional[Callable] = None,
):
    """Decorator for adding implicit differentiation to a root solver.

    Args:
      residual_fun: A callable the returns the possibly nonlinear residual of which to find the root of,
        ``residual_fun(dofs, *args)``.
        The invariant is ``residual_fun(sol, *args) == 0`` at the solution / root ``sol``.
      mat_fun: A callable that returns the sparse tangent matrix as a jax.experimental.BCOO with dofs and
        args as arguments. Can also be a pure callback.
      solve: A linear solver of the form ``solve(mat[jax.experimental.BCOO], b[jnp.ndarray])``.
      free_dofs: For constraining certain degrees of freedom. In case free_dofs is not None, the second argument of the solver
        has to be a dictionary having the keys 'dirichlet dofs' and 'dirichlet conditions'. The first one is a
        dictionary of jnp.ndarrays with the same structure as dofs, where the entries are boolean masks indicating
        the dofs that are constrained. The second one is a dictionary of jnp.ndarrays with the same structure as dofs,
        where the entries are the values of the constrained dofs (see source code of _root_jvp and _root_vjp
        for details or solver.adaptive_load_stepping for exemplary use).
      has_aux: whether the decorated root solver function returns auxiliary data.
      mode: The differentiation mode ('forward' or 'reverse'/'backward').
      reference_signature: optional function signature
        (i.e. arguments and keyword arguments), with which the
        solver and optimality functions are expected to agree. Defaults
        to ``residual_fun``. It is required that solver and optimality
        functions share the same input signature, but both might be
        defined in such a way that the signature correspondence is
        ambiguous (e.g. if both accept catch-all ``**kwargs``). To
        satisfy custom_root's requirement, any function with an
        unambiguous signature can be provided here.

    Returns:
      The decorated root solver function that is equipped a with custom vjp or jvp rule.

    Example:
      See e.g. the implementation of autopdex.solver.adaptive_load_stepping.
    """

    def wrapper(solver_fun):
        return _custom_root(
            solver_fun,
            residual_fun,
            mat_fun,
            free_dofs,
            solve,
            has_aux,
            mode,
            reference_signature,
        )

    return wrapper
